#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from .AdaptivityBase import Builder
from TransportMaps.Routines import L2_misfit

__all__ = ['L2RegressionBuilder',
           'ToleranceSequentialL2RegressionBuilder']

class L2RegressionBuilder(Builder):
    r""" Basis builder through :math:`\mathcal{L}^2` regression

    Given a map :math:`M`, fit a supplied parametric transport map :math:`T`
    through the solution of the :math:`\mathcal{L}^2` regression problem

    .. math::

       \arg\min_{\bf a}\left\Vert M - T \right\Vert_{\mathcal{L}^2} + \alpha \Vert {\bf a} \Vert_x

    where :math:`\alpha\Vert{\bf a}\Vert_x` is a regularization term, with
    respect to one of the available norms.

    Args:
      target_map (:class:`Map<TransportMaps.Maps.Map>`): map :math:`M`
    """
    def __init__(self, target_map=None):
        self.target_map = target_map
        super(L2RegressionBuilder, self).__init__()

    @property
    def target_map(self):
        return self._target_map

    @target_map.setter
    def target_map(self, m):
        self._target_map = m

    def solve(self, transport_map, d=None, qtype=None, qparams=None,
              x=None, w=None, x0=None, regularization=None,
              tol=1e-4, maxit=100, batch_size_list=None,
              mpi_pool_list=None, import_set=set()):
        log_list = transport_map.regression(
            self.target_map, d=d, qtype=qtype, qparams=qparams,
            x=x, w=w, x0=x0, regularization=regularization,
            tol=tol, maxit=maxit, batch_size_list=batch_size_list,
            mpi_pool_list=mpi_pool_list, import_set=import_set)
        return transport_map, log_list

class ToleranceSequentialL2RegressionBuilder(L2RegressionBuilder):
    def solve(self, transport_map_list, eps, regression_params_list, monitor_params):
        tm = None
        x0 = None
        for tm_new, regression_params in zip(
                transport_map_list, regression_params_list):
            if tm is not None:
                for c1, c2 in zip(tm.approx_list, tm_new.approx_list):
                    # Constant part
                    for i1, midx1 in enumerate(c1.c.get_multi_idxs()):
                        for i2, midx2 in enumerate(c2.c.get_multi_idxs()):
                            if midx1 == midx2:
                                break
                        c2.c.coeffs[i2] = c1.c.coeffs[i1]
                    # Integrated part
                    for i1, midx1 in enumerate(c1.h.get_multi_idxs()):
                        for i2, midx2 in enumerate(c2.h.get_multi_idxs()):
                            if midx1 == midx2:
                                break
                        c2.h.coeffs[i2] = c1.h.coeffs[i1]
                x0 = tm_new.coeffs
            tm_new, log_list = super(ToleranceSequentialL2RegressionBuilder, self).solve(
                tm_new, x0=x0, **regression_params)
            if not log_list[-1]['success']:
                if tm is None: # If no map is available return the target map
                    return self.target_map, log_list
                else:
                    break
            else:
                tm = tm_new
            # Check error
            err = L2_misfit(self.target_map, tm, **monitor_params)
            if err < eps:
                break
        return tm, log_list
            


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class XML_Map_unittest(unittest.TestCase):
    def setUp(self):
        self.reg = {'type': 'L2',
                    'alpha': 1e-2}
        self.tol = 1e-3
        self.qtype = 0
        self.qparams = 100
        self.tm = self.load_map()
    
    def load_map(self):
        import TransportMaps.XML as TMXML
        tm = TMXML.load_xml( self.map_fname )
        return tm

    def test_xml_solve(self):
        import numpy as np
        import numpy.random as npr
        import TransportMaps.Distributions as DIST
        tm = self.tm
        tar_mu = npr.randn(tm.dim)
        tar_sig = npr.randn(tm.dim**2).reshape((tm.dim, tm.dim))
        tar_sig2 = np.dot(tar_sig, tar_sig.T)
        target = DIST.GaussianDistribution(tar_mu, sigma=tar_sig2)
        base = DIST.StandardNormalDistribution(tm.dim)
        # Create push forward
        push = DIST.PushForwardTransportMapDistribution(tm, base)
        # Solve
        log = push.minimize_kl_divergence(
            target, qtype=self.qtype, qparams=self.qparams)
        
def build_suite(ttype='all'):
    import os
    suites_list = []
    if ttype in ['all','serial']:
        test_map_xml_dir = os.path.dirname(os.path.realpath(__file__)) + \
                           '/xml/maps/'
        xml_map_list = os.listdir(test_map_xml_dir)
        for xml_map_fname in xml_map_list:
            if xml_map_fname.endswith('.xml'):
                map_name = xml_map_fname.split('.xml')[0]
                test_class = type('XML_Map_' + map_name,
                                  (XML_Map_unittest,),
                                  {'map_fname': test_map_xml_dir + xml_map_fname})
                # Append new suite
                suite = unittest.TestLoader().loadTestsFromTestCase( test_class )
                suites_list.append( suite )
    all_suites = unittest.TestSuite(suites_list)
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()



#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys, getopt
import numpy as np

import TransportMaps.Distributions.Examples.StochasticVolatility as SV

def usage():
    print('DataGeneration.py --output=<filename> [ ' + \
          '-n <max> --phi=.95 --sigma=.25 --mu=-0.5 --input-data=<fname>]')
    print('If no value for phi, sigma or mu are provided, they are considered ' + \
          'hyperparameters')

def full_usage():
    usage()

argv = sys.argv[1:]
nsteps = None
phi = None
sigma = None
mu = None
durbin_data = False
OUT_FNAME = None
IN_FNAME = None
try:
    opts, args = getopt.getopt(argv,"hn:",["output=", "phi=", "sigma=", "mu=",
                                           "input-data="])
except getopt.GetoptError:
    full_usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        full_usage()
        sys.exit()
    elif opt == '-n':
        nsteps = int(arg)
    elif opt == "--phi":
        phi = float(arg)
    elif opt == "--sigma":
        sigma = float(arg)
    elif opt == "--mu":
        mu = float(arg)
    elif opt == "--input-data":
        IN_FNAME = arg
    elif opt == "--output":
        OUT_FNAME = arg
if None in [OUT_FNAME]:
    full_usage()
    sys.exit(3)

#########################################
dataObs = None
if IN_FNAME is not None:
    dataObs = np.loadtxt(IN_FNAME, skiprows=2, usecols=2, delimiter=',')
    if nsteps is None:
        nsteps = len(dataObs)
    else:
        dataObs = dataObs[:nsteps]
    Xt = [None] * nsteps
else:
    dataObs, Xt = SV.generate_data(nsteps, -.5, .25, .95)
    
dens = SV.StocVolHyperDistribution(
    mu is None, sigma is None, phi is None,
    mu=mu, sigma=sigma, phi=phi,
    mu_sigma=1.,
    sigma_k=1., sigma_theta=0.1,
    phi_mean=3., phi_std=1.)

for n in range(nsteps):
    dens.assimilate(y=dataObs[n], Xt=Xt[n])

dens.store(OUT_FNAME)



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from TransportMaps.Distributions.FrozenDistributions import GaussianDistribution
from TransportMaps.Likelihoods.LikelihoodBase import LogLikelihood
from TransportMaps.Distributions.Inference.InferenceBase import BayesPosteriorDistribution
from TransportMaps.Algorithms.SequentialInference.LinearSequentialInference import \
    LinearFilter

from .Coradia175VehicleStateSpace import *

__all__ = ['ParametersLogLikelihood',
           'ParametersPosterior']

class ParametersLogLikelihood(LogLikelihood):
    def __init__(self, y, vehicle, par_name_list):
        self._vehicle = vehicle
        self._par_name_list = par_name_list
        dim = len(par_name_list)
        super(ParametersLogLikelihood, self).__init__(y, dim)

    @property
    def vehicle(self):
        return self._vehicle

    @property
    def par_name_list(self):
        return self._par_name_list

    def evaluate(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("Input dimension mismatch.")
        m = x.shape[0]
        out = np.zeros(m)
        for i in range(m):
            pi_hyper = ParametersPrior(self._par_name_list)
            pi_prior = StateSpacePrior(self.vehicle, self._par_name_list, x[i,:])
            pi_trans = StateSpaceTransition(self.vehicle, self._par_name_list, x[i,:])
            FLT = LinearFilter(pi_hyper=pi_hyper)
            for n, y in enumerate(self.y):
                # Define log-likelihood
                if y is None: # Missing data
                    ll = None
                else: 
                    ll = StateSpaceLogLikelihood(
                        y, self.vehicle, self._par_name_list, init_coeffs=x[i,:])
                # Define transition / prior
                if n > 0: 
                    pin = pi_trans
                else:
                    pin = pi_prior
                # Assimilation
                FLT.assimilate(pin, ll)
            out[i] = FLT.marginal_log_likelihood
        return out

    def grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("Input dimension mismatch.")
        m = x.shape[0]
        out = np.zeros((m,len(self._par_name_list)))
        for i in range(m):
            pi_hyper = ParametersPrior(self._par_name_list)
            pi_prior = StateSpacePrior(self.vehicle, self._par_name_list, x[i,:])
            pi_trans = StateSpaceTransition(self.vehicle, self._par_name_list, x[i,:])
            FLT = LinearFilter(ders=1, pi_hyper=pi_hyper)
            for n, y in enumerate(self.y):
                # Define log-likelihood
                if y is None: # Missing data
                    ll = None
                else: 
                    ll = StateSpaceLogLikelihood(
                        y, self.vehicle, self._par_name_list, init_coeffs=x[i,:])
                # Define transition / prior
                if n > 0: 
                    pin = pi_trans
                else:
                    pin = pi_prior
                # Assimilation
                FLT.assimilate(pin, ll)
            out[i] = FLT.grad_marginal_log_likelihood
        return out

    def tuple_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("Input dimension mismatch.")
        m = x.shape[0]
        ev = np.zeros(m)
        gx = np.zeros((m,len(self._par_name_list)))
        for i in range(m):
            pi_hyper = ParametersPrior(self._par_name_list)
            pi_prior = StateSpacePrior(self.vehicle, self._par_name_list, x[i,:])
            pi_trans = StateSpaceTransition(self.vehicle, self._par_name_list, x[i,:])
            FLT = LinearFilter(ders=1, pi_hyper=pi_hyper)
            for n, y in enumerate(self.y):
                # Define log-likelihood
                if y is None: # Missing data
                    ll = None
                else: 
                    ll = StateSpaceLogLikelihood(
                        y, self.vehicle, self._par_name_list, init_coeffs=x[i,:])
                # Define transition / prior
                if n > 0: 
                    pin = pi_trans
                else:
                    pin = pi_prior
                # Assimilation
                FLT.assimilate(pin, ll)
            ev[i] = FLT.marginal_log_likelihood
            gx[i] = FLT.grad_marginal_log_likelihood
        return (ev, gx)

class ParametersPosterior(BayesPosteriorDistribution):
    def __init__(self, y, vehicle, par_name_list, T=None, Z=None):
        self.T = T
        self.Z = Z
        prior = ParametersPrior(par_name_list)
        logL = ParametersLogLikelihood(y, vehicle, par_name_list)
        super(ParametersPosterior,self).__init__(logL, prior)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest
import numpy as np
import numpy.random as npr
import TransportMaps as TM
import TransportMaps.Distributions as DIST
import TransportMaps.FiniteDifference as FD
import TransportMaps.Functionals as FUNC
import SpectralToolbox.Spectral1D as S1D

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class L2_misfit_DerivativeChecks(unittest.TestCase):

    def setUp(self):
        npr.seed(1)
        self.d = DIST.StandardNormalDistribution(2)
        self.order = 2
        self.fd_eps = 1e-4
        self.qtype = 3
        self.qparams = [10] * 2

    def test_grad_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        def L2_misfit(a, params):
            f1.coeffs = a
            out = TM.L2squared_misfit(f1, f2, d=d, qtype=qtype, qparams=qparams,
                               mpi_pool=mpi_pool)
            return out

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, d=d, qtype=qtype, qparams=qparams,
                                      mpi_pool=mpi_pool)
            return out

        flag = FD.check_grad_a(L2_misfit, grad_a_L2_misfit,
                               self.coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, d=d, qtype=qtype, qparams=qparams,
                                      mpi_pool=mpi_pool)
            return out

        def hess_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.hess_a_L2squared_misfit(f1, f2, d=d, qtype=qtype, qparams=qparams,
                                      mpi_pool=mpi_pool)
            return out

        flag = FD.check_hess_a_from_grad_a(
            grad_a_L2_misfit, hess_a_L2_misfit,
            self.coeffs, self.fd_eps,
            None, verbose=False)
        self.assertTrue( flag )

    def test_action_storage_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        v = np.random.randn( len(self.coeffs) )

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, d=d, qtype=qtype, qparams=qparams,
                                      mpi_pool=mpi_pool)
            return out

        def action_storage_hess_a_L2_misfit(a, v, params):
            f1.coeffs = a
            params1 = {}
            (_, H) = TM.storage_hess_a_L2squared_misfit(
                f1, f2, d=d, params1=params1, qtype=qtype, qparams=qparams,
                mpi_pool=mpi_pool)
            out = TM.action_hess_a_L2squared_misfit(H, v)
            return out

        flag = FD.check_action_hess_a_from_grad_a(
            grad_a_L2_misfit, action_storage_hess_a_L2_misfit,
            self.coeffs, self.fd_eps, v,
            None, verbose=False)
        self.assertTrue( flag )

    def test_batch_grad_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        batch_size = 3
        (x,w) = d.quadrature(qtype,qparams)

        def L2_misfit(a, params):
            f1.coeffs = a
            out = TM.L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                               mpi_pool=mpi_pool)
            return out

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      mpi_pool=mpi_pool)
            return out

        flag = FD.check_grad_a(L2_misfit, grad_a_L2_misfit,
                               self.coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_batch_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        batch_size = 3
        (x,w) = d.quadrature(qtype,qparams)

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      mpi_pool=mpi_pool)
            return out

        def hess_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.hess_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      mpi_pool=mpi_pool)
            return out

        flag = FD.check_hess_a_from_grad_a(
            grad_a_L2_misfit, hess_a_L2_misfit,
            self.coeffs, self.fd_eps,
            None, verbose=False)
        self.assertTrue( flag )

    def test_batch_action_storage_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        v = np.random.randn( len(self.coeffs) )
        batch_size = 3
        (x,w) = d.quadrature(qtype,qparams)

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      mpi_pool=mpi_pool)
            return out

        def action_storage_hess_a_L2_misfit(a, v, params):
            f1.coeffs = a
            (_, H) = TM.storage_hess_a_L2squared_misfit(
                f1, f2, d=d, qtype=qtype, qparams=qparams,
                batch_size=batch_size, mpi_pool=mpi_pool)
            out = TM.action_hess_a_L2squared_misfit(H, v)
            return out

        flag = FD.check_action_hess_a_from_grad_a(
            grad_a_L2_misfit, action_storage_hess_a_L2_misfit,
            self.coeffs, self.fd_eps, v,
            None, verbose=False)
        self.assertTrue( flag )

    def test_precomp_grad_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        (x, w) = d.quadrature(qtype, qparams)
        (_, params1) = TM.mpi_eval('init_mem_precomp_regression',
                                   dmem_key_out_list=['params1'],
                                   obj=f1, mpi_pool=mpi_pool,
                                   concatenate=False)
        dmem_key_in_list = ['params1']
        dmem_arg_in_list = ['precomp']
        dmem_val_in_list = [params1]
        scatter_tuple = (['x'], [x])
        TM.mpi_eval('precomp_regression', scatter_tuple=scatter_tuple,
                    dmem_key_in_list=dmem_key_in_list,
                    dmem_arg_in_list=dmem_arg_in_list,
                    dmem_val_in_list=dmem_val_in_list,
                    obj=f1, mpi_pool=mpi_pool, concatenate=False)

        def L2_misfit(a, params):
            f1.coeffs = a
            out = TM.L2squared_misfit(f1, f2, x=x, w=w, params1=params1,
                               mpi_pool=mpi_pool)
            return out

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, params1=params1,
                                      mpi_pool=mpi_pool)
            return out

        flag = FD.check_grad_a(L2_misfit, grad_a_L2_misfit,
                               self.coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_precomp_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        (x, w) = d.quadrature(qtype, qparams)
        (_, params1) = TM.mpi_eval('init_mem_precomp_regression',
                                   dmem_key_out_list=['params1'],
                                   obj=f1, mpi_pool=mpi_pool,
                                   concatenate=False)
        dmem_key_in_list = ['params1']
        dmem_arg_in_list = ['precomp']
        dmem_val_in_list = [params1]
        scatter_tuple = (['x'], [x])
        TM.mpi_eval('precomp_regression', scatter_tuple=scatter_tuple,
                    dmem_key_in_list=dmem_key_in_list,
                    dmem_arg_in_list=dmem_arg_in_list,
                    dmem_val_in_list=dmem_val_in_list,
                    obj=f1, mpi_pool=mpi_pool, concatenate=False)

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, params1=params1,
                                      mpi_pool=mpi_pool)
            return out

        def hess_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.hess_a_L2squared_misfit(f1, f2, x=x, w=w, params1=params1,
                                      mpi_pool=mpi_pool)
            return out

        flag = FD.check_hess_a_from_grad_a(
            grad_a_L2_misfit, hess_a_L2_misfit,
            self.coeffs, self.fd_eps,
            None, verbose=False)
        self.assertTrue( flag )

    def test_precomp_action_storage_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        (x, w) = d.quadrature(qtype, qparams)
        (_, params1) = TM.mpi_eval('init_mem_precomp_regression',
                                   dmem_key_out_list=['params1'],
                                   obj=f1, mpi_pool=mpi_pool,
                                   concatenate=False)
        dmem_key_in_list = ['params1']
        dmem_arg_in_list = ['precomp']
        dmem_val_in_list = [params1]
        scatter_tuple = (['x'], [x])
        TM.mpi_eval('precomp_regression', scatter_tuple=scatter_tuple,
                    dmem_key_in_list=dmem_key_in_list,
                    dmem_arg_in_list=dmem_arg_in_list,
                    dmem_val_in_list=dmem_val_in_list,
                    obj=f1, mpi_pool=mpi_pool, concatenate=False)
        v = np.random.randn( len(self.coeffs) )

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, params1=params1,
                                      mpi_pool=mpi_pool)
            return out

        def action_storage_hess_a_L2_misfit(a, v, params):
            f1.coeffs = a
            (_, H) = TM.storage_hess_a_L2squared_misfit(
                f1, f2, x=x, w=w, params1=params1,
                mpi_pool=mpi_pool)
            out = TM.action_hess_a_L2squared_misfit(H, v)
            return out

        flag = FD.check_action_hess_a_from_grad_a(
            grad_a_L2_misfit, action_storage_hess_a_L2_misfit,
            self.coeffs, self.fd_eps, v,
            None, verbose=False)
        self.assertTrue( flag )

    def test_precomp_batch_grad_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        batch_size = 3
        (x,w) = d.quadrature(qtype,qparams)
        (_, params1) = TM.mpi_eval('init_mem_precomp_regression',
                                   dmem_key_out_list=['params1'],
                                   obj=f1, mpi_pool=mpi_pool,
                                   concatenate=False)
        dmem_key_in_list = ['params1']
        dmem_arg_in_list = ['precomp']
        dmem_val_in_list = [params1]
        scatter_tuple = (['x'], [x])
        TM.mpi_eval('precomp_regression', scatter_tuple=scatter_tuple,
                    dmem_key_in_list=dmem_key_in_list,
                    dmem_arg_in_list=dmem_arg_in_list,
                    dmem_val_in_list=dmem_val_in_list,
                    obj=f1, mpi_pool=mpi_pool, concatenate=False)

        def L2_misfit(a, params):
            f1.coeffs = a
            out = TM.L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                               params1=params1, mpi_pool=mpi_pool)
            return out

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      params1=params1, mpi_pool=mpi_pool)
            return out

        flag = FD.check_grad_a(L2_misfit, grad_a_L2_misfit,
                               self.coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_precomp_batch_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        batch_size = 3
        (x,w) = d.quadrature(qtype,qparams)
        (_, params1) = TM.mpi_eval('init_mem_precomp_regression',
                                   dmem_key_out_list=['params1'],
                                   obj=f1, mpi_pool=mpi_pool,
                                   concatenate=False)
        dmem_key_in_list = ['params1']
        dmem_arg_in_list = ['precomp']
        dmem_val_in_list = [params1]
        scatter_tuple = (['x'], [x])
        TM.mpi_eval('precomp_regression', scatter_tuple=scatter_tuple,
                    dmem_key_in_list=dmem_key_in_list,
                    dmem_arg_in_list=dmem_arg_in_list,
                    dmem_val_in_list=dmem_val_in_list,
                    obj=f1, mpi_pool=mpi_pool, concatenate=False)

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      params1=params1, mpi_pool=mpi_pool)
            return out

        def hess_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.hess_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      params1=params1, mpi_pool=mpi_pool)
            return out

        flag = FD.check_hess_a_from_grad_a(
            grad_a_L2_misfit, hess_a_L2_misfit,
            self.coeffs, self.fd_eps,
            None, verbose=False)
        self.assertTrue( flag )

    def test_precomp_batch_action_storage_hess_a_l2_misfit(self):
        d = self.d
        f1 = self.f1
        f2 = self.f2
        qtype = self.qtype
        qparams = self.qparams
        mpi_pool = self.mpi_pool

        v = np.random.randn( len(self.coeffs) )
        batch_size = 3
        (x,w) = d.quadrature(qtype,qparams)
        (_, params1) = TM.mpi_eval('init_mem_precomp_regression',
                                   dmem_key_out_list=['params1'],
                                   obj=f1, mpi_pool=mpi_pool,
                                   concatenate=False)
        dmem_key_in_list = ['params1']
        dmem_arg_in_list = ['precomp']
        dmem_val_in_list = [params1]
        scatter_tuple = (['x'], [x])
        TM.mpi_eval('precomp_regression', scatter_tuple=scatter_tuple,
                    dmem_key_in_list=dmem_key_in_list,
                    dmem_arg_in_list=dmem_arg_in_list,
                    dmem_val_in_list=dmem_val_in_list,
                    obj=f1, mpi_pool=mpi_pool, concatenate=False)

        def grad_a_L2_misfit(a, params):
            f1.coeffs = a
            out = TM.grad_a_L2squared_misfit(f1, f2, x=x, w=w, batch_size=batch_size,
                                      params1=params1, mpi_pool=mpi_pool)
            return out

        def action_storage_hess_a_L2_misfit(a, v, params):
            f1.coeffs = a
            (_, H) = TM.storage_hess_a_L2squared_misfit(
                f1, f2, x=x, w=w, batch_size=batch_size,
                params1=params1, mpi_pool=mpi_pool)
            out = TM.action_hess_a_L2squared_misfit(H, v)
            return out

        flag = FD.check_action_hess_a_from_grad_a(
            grad_a_L2_misfit, action_storage_hess_a_L2_misfit,
            self.coeffs, self.fd_eps, v,
            None, verbose=False)
        self.assertTrue( flag )

#
# Serial and parallel tests
#
class Serial_L2_misfit_DerivativeChecks(L2_misfit_DerivativeChecks):
    def setUp(self):
        super(Serial_L2_misfit_DerivativeChecks,self).setUp()
        self.mpi_pool = None

class Parallel_L2_misfit_DerivativeChecks(L2_misfit_DerivativeChecks):
    def setUp(self):
        super(Parallel_L2_misfit_DerivativeChecks,self).setUp()
        import_set = set([ ('numpy', 'np') ])
        self.mpi_pool = TM.get_mpi_pool()
        self.mpi_pool.start(2, import_set)
    def tearDown(self):
        import time
        self.mpi_pool.stop()
        time.sleep(0.2)

#
# Linear Span and Integrated Exponential and Integrated Squared
#
class LinSpan():
    def setUp_approx(self):
        basis_list = [ S1D.HermiteProbabilistsPolynomial() ] * 2
        order_list = [ self.order ]*2
        # 2d Random target function
        self.f2 = FUNC.LinearSpanApproximation(
            basis_list, spantype='total', order_list=order_list)
        self.f2.coeffs = np.random.randn( self.f2.n_coeffs ) 
        # Approximating Function
        self.f1 = FUNC.LinearSpanApproximation(
            basis_list, spantype='total', order_list=order_list)
        # Random point on coeff space
        self.coeffs = np.random.randn( self.f1.n_coeffs )
class MonoLinSpan():
    def setUp_approx(self):
        basis_list = [ S1D.HermiteProbabilistsPolynomial() ] * 2
        order_list = [ self.order ]*2
        # 2d Random target function
        self.f2 = FUNC.MonotonicLinearSpanApproximation(
            basis_list, spantype='total', order_list=order_list)
        self.f2.coeffs = np.random.randn( self.f2.n_coeffs ) 
        # Approximating Function
        self.f1 = FUNC.MonotonicLinearSpanApproximation(
            basis_list, spantype='total', order_list=order_list)
        # Random point on coeff space
        self.coeffs = np.random.randn( self.f1.n_coeffs )
class IntExp():
    def setUp_approx(self):
        c_basis_list = [ S1D.HermiteProbabilistsPolynomial() ] * 2
        e_basis_list = [ S1D.ConstantExtendedHermiteProbabilistsFunction() ] * 2
        c_order_list = [self.order, 0]
        e_order_list = [self.order, self.order-1]
        # 2d Random target function
        c2 = FUNC.LinearSpanApproximation(c_basis_list, spantype='total', order_list=c_order_list)
        e2 = FUNC.LinearSpanApproximation(e_basis_list, spantype='total', order_list=e_order_list)
        self.f2 = FUNC.MonotonicIntegratedExponentialApproximation(c2, e2)
        self.f2.coeffs = np.random.randn( self.f2.n_coeffs ) 
        # Approximating Function
        c1 = FUNC.LinearSpanApproximation(c_basis_list, spantype='total', order_list=c_order_list)
        e1 = FUNC.LinearSpanApproximation(e_basis_list, spantype='total', order_list=e_order_list)
        self.f1 = FUNC.MonotonicIntegratedExponentialApproximation(c1, e1)
        # Random point on coeff space
        self.coeffs = np.random.randn( self.f1.n_coeffs )
class IntSq():
    def setUp_approx(self):
        c_basis_list = [ S1D.HermiteProbabilistsPolynomial() ] * 2
        e_basis_list = [ S1D.HermiteProbabilistsPolynomial() ] * 2
        c_order_list = [self.order, 0]
        e_order_list = [self.order, self.order-1]
        # 2d Random target function
        c2 = FUNC.LinearSpanApproximation(c_basis_list, spantype='total', order_list=c_order_list)
        e2 = FUNC.LinearSpanApproximation(e_basis_list, spantype='total', order_list=e_order_list)
        self.f2 = FUNC.MonotonicIntegratedSquaredApproximation(c2, e2)
        self.f2.coeffs = np.random.randn( self.f2.n_coeffs ) 
        # Approximating Function
        c1 = FUNC.LinearSpanApproximation(c_basis_list, spantype='total', order_list=c_order_list)
        e1 = FUNC.LinearSpanApproximation(e_basis_list, spantype='total', order_list=e_order_list)
        self.f1 = FUNC.MonotonicIntegratedSquaredApproximation(c1, e1)
        # Random point on coeff space
        self.coeffs = np.random.randn( self.f1.n_coeffs )

        
#
# Single tests
#
class Serial_LinSpan_L2_misfit_DerivativeChecks(
        Serial_L2_misfit_DerivativeChecks, LinSpan):
    def setUp(self):
        super(Serial_LinSpan_L2_misfit_DerivativeChecks,self).setUp()
        super(Serial_LinSpan_L2_misfit_DerivativeChecks,self).setUp_approx()
class Serial_MonoLinSpan_L2_misfit_DerivativeChecks(
        Serial_L2_misfit_DerivativeChecks, MonoLinSpan):
    def setUp(self):
        super(Serial_MonoLinSpan_L2_misfit_DerivativeChecks,self).setUp()
        super(Serial_MonoLinSpan_L2_misfit_DerivativeChecks,self).setUp_approx()
class Serial_IntExp_L2_misfit_DerivativeChecks(
        Serial_L2_misfit_DerivativeChecks, IntExp):
    def setUp(self):
        super(Serial_IntExp_L2_misfit_DerivativeChecks,self).setUp()
        super(Serial_IntExp_L2_misfit_DerivativeChecks,self).setUp_approx()
class Serial_IntSq_L2_misfit_DerivativeChecks(
        Serial_L2_misfit_DerivativeChecks, IntSq):
    def setUp(self):
        super(Serial_IntSq_L2_misfit_DerivativeChecks,self).setUp()
        super(Serial_IntSq_L2_misfit_DerivativeChecks,self).setUp_approx()
class Parallel_LinSpan_L2_misfit_DerivativeChecks(
        Parallel_L2_misfit_DerivativeChecks, LinSpan):
    def setUp(self):
        super(Parallel_LinSpan_L2_misfit_DerivativeChecks,self).setUp()
        super(Parallel_LinSpan_L2_misfit_DerivativeChecks,self).setUp_approx()
class Parallel_MonoLinSpan_L2_misfit_DerivativeChecks(
        Parallel_L2_misfit_DerivativeChecks, MonoLinSpan):
    def setUp(self):
        super(Parallel_MonoLinSpan_L2_misfit_DerivativeChecks,self).setUp()
        super(Parallel_MonoLinSpan_L2_misfit_DerivativeChecks,self).setUp_approx()
class Parallel_IntExp_L2_misfit_DerivativeChecks(
        Parallel_L2_misfit_DerivativeChecks, IntExp):
    def setUp(self):
        super(Parallel_IntExp_L2_misfit_DerivativeChecks,self).setUp()
        super(Parallel_IntExp_L2_misfit_DerivativeChecks,self).setUp_approx()
class Parallel_IntSq_L2_misfit_DerivativeChecks(
        Parallel_L2_misfit_DerivativeChecks, IntSq):
    def setUp(self):
        super(Parallel_IntSq_L2_misfit_DerivativeChecks,self).setUp()
        super(Parallel_IntSq_L2_misfit_DerivativeChecks,self).setUp_approx()
        
def build_suite(ttype='all'):
    suite_se_ls_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Serial_LinSpan_L2_misfit_DerivativeChecks )
    suite_se_mls_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Serial_MonoLinSpan_L2_misfit_DerivativeChecks )
    suite_se_ie_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Serial_IntExp_L2_misfit_DerivativeChecks )
    suite_se_sq_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Serial_IntSq_L2_misfit_DerivativeChecks )
    suite_pa_ls_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_LinSpan_L2_misfit_DerivativeChecks )
    suite_pa_mls_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_MonoLinSpan_L2_misfit_DerivativeChecks )
    suite_pa_ie_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_IntExp_L2_misfit_DerivativeChecks )
    suite_pa_sq_l2_mf = unittest.TestLoader().loadTestsFromTestCase(
        Parallel_IntSq_L2_misfit_DerivativeChecks )

    # Serial
    suites_list = []
    if ttype in ['all','serial']:
        suites_list = [
            suite_se_ls_l2_mf,
            suite_se_mls_l2_mf, 
            suite_se_ie_l2_mf,
            suite_se_sq_l2_mf
        ]
    # Parallel
    if ttype in ['all','parallel'] and MPI_SUPPORT:
        suites_list += [
            suite_pa_ls_l2_mf,
            suite_pa_mls_l2_mf,
            suite_pa_ie_l2_mf,
            suite_pa_sq_l2_mf
        ]

    all_suites = unittest.TestSuite( suites_list )
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)
    
if __name__ == '__main__':
    run_tests()


#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

__all__ = ['AVAIL_MONOTONE', 'AVAIL_SPAN', 'AVAIL_BTYPE',
           'AVAIL_QTYPE', 'AVAIL_DERS',
           'AVAIL_ADAPTIVITY', 'AVAIL_REGRESSION_ADAPTIVITY',
           'AVAIL_MCMC_ALGORITHMS', 'AVAIL_LOGGING',
           'AVAIL_DISTRIBUTIONS',
           'print_avail_options']

from TransportMaps.External import *

# MONOTONE APPROXIMATION
# 'intexp': Integrated Exponential
# 'linspan': Constrained Polynomial
AVAIL_MONOTONE = {'linspan': 'monotone linear span',
                  'intexp': 'integrated exponential',
                  'intsq': 'integrated square'}

# SPAN APPROXIMATION
# 'full': Full order approximation
# 'total': Total order approximation
AVAIL_SPAN = {'full': 'full span',
              'total': 'total order span'}

# BASIS TYPES
# 'poly': Hermite polynomials
# 'rbf': Radial basis functions
AVAIL_BTYPE = {'poly': 'polynomial basis',
               'rbf': 'radial basis functions (requires SPAN=full)'}

# QUADRATURES
# 0: Monte Carlo
# 3: Gauss quadrature
AVAIL_QTYPE = {0: 'Monte Carlo',
               3: 'Gauss quadrature'}

# DERS
# 1: BFGS
# 2: Newton-CG
AVAIL_DERS = {0: 'BFGS (gradient free)',
              1: 'BFGS (gradient needed)',
              2: 'Newton-CG (Hessian needed)'}

# ADAPTIVITY
# 'none': no adaptivity performed
# 'sequential': a sequence of maps of increasing order are used
AVAIL_ADAPTIVITY = {
    'none': 'no adaptivity',
    'sequential': 'a prefix sequence of maps is used',
    'tol-sequential': 'a sequence of maps is used until tolerance is met',
}

# REGRESSION ADAPTIVITY
# 'none': no adaptivity performed
# 'tol-sequential': a sequence of maps of increasing order are used
AVAIL_REGRESSION_ADAPTIVITY = {
    'none': 'no adaptivity',
    'tol-sequential': 'meet a tolerance using a prefix sequence of maps'}

# MCMC ALGORITHMS
AVAIL_MCMC_ALGORITHMS = {
    'mhind': 'Metropoli-Hastings with independent proposals',
    'hmc': "Hamiltonian Monte Carlo",
    }

# LOGGING
AVAIL_LOGGING = {10: 'debug',
                 20: 'info',
                 30: 'warning',
                 40: 'error',
                 50: 'critical'}

AVAIL_DISTRIBUTIONS = {'exact-target': 'exact target density',
                       'approx-target': 'approximate target density',
                       'exact-base': 'exact base density',
                       'approx-base': 'approximate base density'}

def print_avail_options(avail, prefix='', header=True):
    out_str = ''
    if header:
        out_str += prefix + 'Available options:\n'
    for key, val in sorted(avail.items()):
        out_str += prefix + '  ' + str(key) + ': ' + val + '\n'
    return out_str



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import time
import numpy as np

from TransportMaps import mpi_eval

__all__ = ['compute_vals_variance_approx_kl', 'variance_approx_kl']

def compute_vals_variance_approx_kl( d1, d2, params1=None, params2=None, x=None,
                                     mpi_pool_tuple=(None,None), import_set=set() ):
    r""" Compute values necessary for the evaluation of the variance diagnostic :func:`variance_approx_kl`

    Returns:
      (:class:`tuple` [2] :class:`ndarray<numpy.ndarray>` [:math:`m`]) --
        computed values of :math:`\log\pi_1` and :math:`\log\pi_2`

    .. seealso:: :func:`variance_approx_kl`
    """
    # d1
    scatter_tuple = (['x'], [x])
    vals_d1 = mpi_eval("log_pdf", obj=d1, scatter_tuple=scatter_tuple,
                       mpi_pool=mpi_pool_tuple[0])
    # d2
    vals_d2 = mpi_eval("log_pdf", obj=d2, scatter_tuple=scatter_tuple,
                       mpi_pool=mpi_pool_tuple[1])
    return (vals_d1, vals_d2)

def variance_approx_kl( d1, d2, params1=None, params2=None, vals_d1=None, vals_d2=None,
                        qtype=None, qparams=None, x=None, w=None,
                        mpi_pool_tuple=(None,None), import_set=set() ):
    r""" Variance diagnositc

    Statistical analysis of the variance diagnostic

    .. math::

       \mathcal{D}_{KL}(\pi_1 \Vert \pi_2) \approx \frac{1}{2} \mathbb{V}_{\pi_1} \left( \log \frac{\pi_1}{\pi_2}\right)

    Args:
      d1 (Distribution): distribution :math:`\pi_1`
      d2 (Distribution): distribution :math:`\pi_2`
      params1 (dict): parameters for distribution :math:`\pi_1`
      params2 (dict): parameters for distribution :math:`\pi_2`
      vals_d1 (:class:`ndarray<numpy.ndarray>` [:math:`m`]):
        computed values of :math:`\log\pi_1`
      vals_d2 (:class:`ndarray<numpy.ndarray>` [:math:`m`]):
        computed values of:math:`\log\pi_2`
      qtype (int): quadrature type to be used for the approximation of
        :math:`\mathbb{E}_{\pi_1}`
      qparams (object): parameters necessary for the construction of the
        quadrature
      x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): quadrature points
        used for the approximation of :math:`\mathbb{E}_{\pi_1}`
      w (:class:`ndarray<numpy.ndarray>` [:math:`m`]): quadrature weights
        used for the approximation of :math:`\mathbb{E}_{\pi_1}`
      mpi_pool_tuple (:class:`tuple` [2] of :class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`):
        pool of processes to be used for the evaluation of ``d1`` and ``d2``
      import_set (set): list of couples ``(module_name,as_field)`` to be imported
        as ``import module_name as as_field`` (for MPI purposes)

    .. note:: The parameters ``(qtype,qparams)`` and ``(x,w)`` are mutually
      exclusive, but one pair of them is necessary.
    """
    if vals_d1 is None or vals_d2 is None:
        if x is None and w is None:
            (x,w) = d1.quadrature(qtype, qparams, mpi_pool=mpi_pool_tuple[0])
        elif x is None or w is None:
            raise ValueError("Provide quadrature points and weights or quadrature " + \
                             "type and parameters")
    if vals_d1 is None or vals_d2 is None:
        vals_d1, vals_d2 = compute_vals_variance_approx_kl(
            d1, d2, params1, params2, x=x,
            mpi_pool_tuple=mpi_pool_tuple, import_set=import_set)
    vals = vals_d1 - vals_d2
    expect = np.dot( vals, w )
    var = .5 * np.dot( (vals - expect)**2., w )
    return var



#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import sys, getopt
import warnings
import time
import numpy as np
import numpy.random as npr
import scipy.stats as stats
from tabulate import tabulate
import time
from datetime import timedelta

try:
    import matplotlib.pyplot as plt
except ImportError:
    PLOT_SUPPORT = False
    warnings.warn("No plotting support")
else:
    PLOT_SUPPORT = True

import TransportMaps as TM
import TransportMaps.Distributions as DIST
import TransportMaps.tests.TestFunctions as TF

# # DEBUG MPI
# import mpi_map
# mpi_map.logger.setLevel(logging.DEBUG)

# DEBUG everything
logging.basicConfig(level=logging.INFO)

# TESTS
#  0: 1D - N(mu,sigma)
#  NOT WORKING 1: 1D - Mixture X~N(-1,1), Y~N(1,0.5), p=0.3
#  2: 1D - T(x) = a + b*x + c * arctan(d + e*x)
#  3: 1D - Y = T(x) = exp(a*x) -- Corresponding to Y~logN(0,a^2)
#  4: 1D - Y~Logistic(\mu,s)
#  5: 1D - Y~Gamma(loc,scale)
#  6: 1D - Y~Beta(a,b)
#  7: 1D - Gumbel distribution
#  8: 1D - Weibull distribution
#  9: 2D - N(mu,sigma)
# 10: 2D - Banana
AVAIL_TESTN = {0: '1D Std. Normal',
               2: '1D ArcTan',
               3: '1D LogNormal',
               4: '1D Logistic',
               5: '1D Gamma',
               6: '1D Beta',
               7: '1D Gumbel',
               8: '1D Weibull',
               9: '2D Std. Normal',
               10: '2D Banana'}

# SPAN APPROXIMATION
# 'full': Full order approximation
# 'total': Total order approximation
AVAIL_SPAN = ['full', 'total']

# QUADRATURES
# 0: Monte Carlo
# 3: Gauss quadrature
AVAIL_QUADRATURE = {0: 'Monte Carlo',
                    3: 'Gauss quadrature'}

# BASIS TYPES
AVAIL_BTYPE = ['poly', 'rbf']

def usage():
    print('python KLdivergence-InverseMap-DirectConstruction.py -t <test_number> -s <span_type> -q <quad_type> -n <quad_n_points> [--btype=<basis_type> --tol=<tolerance> --max-it=<max_iterations> --with-reg=<alpha> --ord-list=<comma_sep_order_list> --no-plotting --no-l2 --nprocs=<nprocs>]')

def print_avail_span():
    print('Available <span_type>: %s' % AVAIL_SPAN)

def print_avail_qtype():
    print('Available <quad_type>:')
    for qtypen, qtypename in AVAIL_QUADRATURE.items():
        print('  %d: %s' % (qtypen, qtypename))

def print_avail_btype():
    print('Available <basis_type>: %s' % AVAIL_BTYPE)

def print_avail_testn():
    print('Available <test_number>:')
    for testn, testname in AVAIL_TESTN.items():
        print('  %d: %s' % (testn, testname))

def print_example():
    print('python KLdivergence-InverseMap-DirectConstruction.py -t 2 -s total -q 3 -n 30 --tol=1e-4 --with-reg=1e-3 --ord-list=1,3,5,7,9')

def full_usage():
    usage()
    print_avail_testn()
    print_avail_span()
    print_avail_qtype()
    print_avail_btype()
    print_example()

argv = sys.argv[1:]

PLOTTING = True
L2_ERR = True
TESTN = None
SPAN = None
QTYPE = None
BTYPE = 'poly'
QNUM = None
REG = None
# Orders
ORDERS = [1,3,5]
# Tolerance
TOL = 1e-4
MAXIT = 200
NPROCS = 1
try:
    opts, args = getopt.getopt(argv,"ht:s:q:n:",["testn=", "span_type=",
                                                 "qtype=", "qnum=", "btype=",
                                                 "tol=", "max-it=",
                                                 "with-reg=", "ord-list=",
                                                 "no-plotting", "no-l2", "nprocs="])
except getopt.GetoptError:
    full_usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        full_usage()
        sys.exit()
    elif opt in ("-t", "--testn"):
        TESTN = int(arg)
        if TESTN not in AVAIL_TESTN:
            usage()
            print_avail_testn()
            sys.exit(1)
    elif opt in ("-s", "--span_type"):
        SPAN = arg
        if SPAN not in AVAIL_SPAN:
            usage()
            print_avail_span()
            sys.exit(1)
    elif opt in ("-q", "--qtype"):
        QTYPE = int(arg)
        if QTYPE not in AVAIL_QUADRATURE:
            usage()
            print_avail_qtype()
            sys.exit(1)
    elif opt in ("--btype"):
        BTYPE = arg
        if BTYPE not in AVAIL_BTYPE:
            usage()
            print_avail_btype()
            sys.exit(1)
    elif opt in ("-n", "--qnum"):
        QNUM = int(arg)
    elif opt in ("--tol"):
        TOL = float(arg)
    elif opt in ("--max-it"):
        MAXIT = int(arg)
    elif opt in ("--with-reg"):
        REG = {'type': 'L2',
               'alpha': float(arg)}
    elif opt in ("--ord-list"):
        ORDERS = [int(s) for s in arg.split(',')]
    elif opt in ("--no-plotting"):
        PLOTTING = False
    elif opt in ("--no-l2"):
        L2_ERR = False
    elif opt in ("--nprocs"):
        NPROCS = int(arg)
if None in [TESTN, SPAN, QTYPE, QNUM]:
    full_usage()
    sys.exit(3)

mpi_pool = None
if NPROCS > 1:
    mpi_pool = TM.get_mpi_pool()
    mpi_pool.start(NPROCS)

try:
    # PLOTTING
    fsize = (6,5)
    discr = 101

    # Get test title and parameters
    title, setup, test = TF.get(TESTN)

    # # REGULARIZATION
    # # None: No regularization
    # # L2: L2 regularization
    # REG = None
    # # REG = {'type': 'None',
    # #        'alpha': 1e-3}
    if REG is None:
        tit_reg = 'None'
    elif REG['type'] == 'L2':
        tit_reg = REG['type'] + '-' + str(REG['alpha'])

    # L2 error estimation
    # 3: Gauss quadrature of order n
    # 0: Monte Carlo quadrature with n point
    qtype = QTYPE
    if qtype == 0:
        qparams = QNUM
        tit_intest = "MC" + '-' + str(qparams)
    elif qtype == 3:
        qparams = [QNUM] * setup['dim']
        tit_intest = "Quad" + '-' + str(qparams[0])

    # Gradient information
    # 0: derivative free
    # 1: gradient information
    # 2: Hessian information
    ders = 1

    # MC samples
    N_samp = 10000

    # Generate samples of reference and target
    ref_samp = test['base_distribution'].rvs(N_samp)
    T_samp = test['support_map']( test['target_map'](ref_samp) )

    # PLOTTING PRE-COMPUTATION
    # Start plotting of the 1d map
    if PLOTTING and PLOT_SUPPORT:
        pspan = setup['plotspan']
        rspan = setup['refplotspan']

        # Plot the 1d map
        X1d = np.vstack( (np.linspace(rspan[0,0],rspan[0,1],discr),
                          np.zeros((setup['dim']-1,discr))) ).T
        t1 = test['support_map']( test['target_map'](X1d) )[:,0]
        fig_map_1d = plt.figure()
        ax_map_1d = fig_map_1d.add_subplot(1,1,1)
        ax_map_1d.plot(X1d[:,0], t1, 'k--', linewidth=2, label='exact')

        # Plot the 1d distribution
        fig_kde_1d = plt.figure()
        ax_kde_1d = fig_kde_1d.add_subplot(1,1,1)
        # Exact distribution 1d
        X1d_kde = np.linspace(pspan[0,0],pspan[0,1],discr).reshape((discr,1))
        if setup['dim'] == 1:
            # Plot analytic
            pdf1d = test['target_distribution'].pdf(X1d_kde)
            ax_kde_1d.plot(X1d_kde, pdf1d, 'k--', linewidth=2, label='$\pi^1_{tar}$')
        else:
            # Gaussian KDE approximation
            T1_kde = stats.gaussian_kde(T_samp[:,0][:,np.newaxis].T)
            pdf1d = T1_kde(X1d_kde.T)
            ax_kde_1d.plot(X1d_kde, pdf1d, 'k--', linewidth=2, label='kde-$\pi^1_{tar}$')

        if setup['dim'] >= 2:
            # Plot the 2d transport map
            x = np.linspace(rspan[0,0],rspan[0,1],discr)
            y = np.linspace(rspan[1,0],rspan[1,1],discr)
            xx,yy = np.meshgrid(x,y)
            X2d = np.vstack( (xx.flatten(),yy.flatten(),np.zeros((setup['dim']-2,discr**2))) ).T
            t2 = test['support_map']( test['target_map'](X2d) )[:,1]
            fig_maps2d = []
            ax_maps = []
            levels_maps2d = np.linspace(np.min(t2),np.max(t2),20)
            for i,order in enumerate(ORDERS):
                fig_maps2d.append( plt.figure() )
                ax_map = fig_maps2d[-1].add_subplot(1,1,1)
                ax_map.contour(xx,yy,t2.reshape(xx.shape),
                               levels=levels_maps2d)
                ax_maps.append( ax_map )
                plt.title("Order %d" % order)

            # Plot PDF 2d
            fig_pdf2d = []
            axs_pdf2d = []
            x = np.linspace(pspan[0,0],pspan[0,1],discr)
            y = np.linspace(pspan[1,0],pspan[1,1],discr)
            xx2d,yy2d = np.meshgrid(x,y)
            X2d_kde = np.vstack( (xx2d.flatten(),yy2d.flatten(),
                                  np.zeros((setup['dim']-2,discr**2))) ).T
            if setup['dim'] == 2:
                pdf2d = test['target_distribution'].pdf(X2d_kde)
            else:
                T2_kde = stats.gaussian_kde(T_samp[:,:1].T)
                pdf2d = T2_kde(X2d_kde.T)
            levels_pdf2d = np.linspace(np.min(pdf2d),np.max(pdf2d),20)
            for i,order in enumerate(ORDERS):
                fig_pdf2d.append( plt.figure() )
                ax_pdf2d = fig_pdf2d[-1].add_subplot(1,1,1)
                ax_pdf2d.contour(xx2d,yy2d,pdf2d.reshape(xx2d.shape),
                                 levels=levels_pdf2d)
                axs_pdf2d.append( ax_pdf2d )
                plt.title("Order %d" % order)

    # Define header for the output log
    tit_type = 'monot'
    if ders in [0,1]:
        log_header = ['Order','#F.ev.','#J.ev']
    else:
        log_header = ['Order','#F.ev.','#J.ev','#H.ev']
    if L2_ERR:
        log_header.append('$L^2$-err')

    log = [ ]

    for i_ord,order in enumerate(ORDERS):
        print("\nOrder: %d" % order)
        # Build the transport map (isotropic for each entry)
        tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            setup['dim'], order, SPAN, btype=BTYPE)
        # tm_approx = TM.Default_IsotropicLinearSpanTriangularTransportMap(
        #     setup['dim'], order, SPAN, btype=BTYPE)

        # Construct distribution
        # Distribution T^\sharp \pi
        tm_distribution = DIST.PullBackTransportMapDistribution(tm_approx, test['base_distribution'])
        # Target distribution to be approximated L^\sharp \pi_{\rm tar}
        target_distribution = DIST.PullBackTransportMapDistribution(test['support_map'],
                                                          test['target_distribution'])
        # SOLVE min D_KL (T^\sharp \pi || L^\sharp \pi_{\rm tar})
        start = time.clock()
        log_entry_solve = tm_distribution.minimize_kl_divergence(target_distribution, qtype=qtype,
                                                            qparams=qparams,
                                                            regularization=REG,
                                                            tol=TOL, maxit=MAXIT,
                                                            ders=ders,
                                                            batch_size=[None,None],
                                                            mpi_pool=mpi_pool)
        stop = time.clock()
        print("Solve time: %s" % timedelta(seconds=(stop-start)))
        log_entry = [order] + log_entry_solve

        # Construct approximate distribution L_\sharp T^\sharp \pi
        approx_distribution = DIST.PushForwardTransportMapDistribution(test['support_map'], tm_distribution)

        if L2_ERR or PLOTTING:
            # Push forward reference samples in two steps
            # (1) Map x from \pi to y from T^\sharp \pi
            # (2) Map y from T^\sharp \pi to z from L_\sharp T^\sharp \pi
            scatter_tuple = (['x'],[ref_samp])
            tmp = TM.mpi_eval("map_samples_base_to_target", scatter_tuple=scatter_tuple,
                              obj=tm_distribution, mpi_pool=mpi_pool)
            scatter_tuple = (['x'],[tmp])
            fapp_samp = TM.mpi_eval("map_samples_base_to_target", scatter_tuple=scatter_tuple,
                                    obj=approx_distribution, mpi_pool=mpi_pool) 

        # PLOT
        if PLOTTING and PLOT_SUPPORT:
            # Map to T^\sharp \pi
            scatter_tuple = (['x'],[X1d])
            t1_approx_tmp = TM.mpi_eval("map_samples_base_to_target",
                                        scatter_tuple=scatter_tuple,
                                        obj=tm_distribution, mpi_pool=mpi_pool)
            # Map to L_\sharp T^\sharp \pi
            scatter_tuple = (['x'],[t1_approx_tmp])
            t1_approx = TM.mpi_eval("map_samples_base_to_target",
                                    scatter_tuple=scatter_tuple,
                                    obj=approx_distribution,
                                    mpi_pool=mpi_pool)[:,0]
            # Plot 1d transport map
            ax_map_1d.plot(X1d[:,0], t1_approx, label='ord %d' % order)

            # Plot the 1d distribution
            if setup['dim'] == 1:
                scatter_tuple = (['x'],[X1d_kde])
                pdf1d_approx = TM.mpi_eval("pdf", scatter_tuple=scatter_tuple,
                                           obj=approx_distribution, mpi_pool=mpi_pool)
            else:
                t1_kde_approx = stats.gaussian_kde(fapp_samp[:,0][:,np.newaxis].T)
                pdf1d_approx = t1_kde_approx(X1d_kde.T)
            ax_kde_1d.plot(X1d_kde, pdf1d_approx, label='ord %d' % order)


            if setup['dim'] >= 2:
                # Map to T^\sharp \pi
                scatter_tuple = (['x'],[X2d])
                t2_approx_tmp = TM.mpi_eval("map_samples_base_to_target",
                                            scatter_tuple=scatter_tuple,
                                            obj=tm_distribution, mpi_pool=mpi_pool)
                # Map to L_\sharp T^\sharp \pi
                scatter_tuple = (['x'],[t2_approx_tmp])
                t2_approx = TM.mpi_eval("map_samples_base_to_target",
                                        scatter_tuple=scatter_tuple,
                                        obj=approx_distribution, mpi_pool=mpi_pool)[:,1]
                # Plot 2d transport map
                ax_maps[i_ord].contour(xx2d, yy2d, t2_approx.reshape(xx2d.shape),
                                       linestyles='dashed', levels=levels_maps2d)

                # Plot the 2d distribution
                if setup['dim'] == 2:
                    scatter_tuple = (['x'],[X2d_kde])
                    pdf2d_approx = TM.mpi_eval("pdf", scatter_tuple=scatter_tuple,
                                               obj=approx_distribution, mpi_pool=mpi_pool)
                else:
                    t2_kde_approx = stats.gaussian_kde(fapp_samp[:,:1].T)
                    pdf2d_approx = t2_kde_approx(X2d_kde.T)
                axs_pdf2d[i_ord].contour(xx2d,yy2d,pdf2d_approx.reshape(xx2d.shape),
                                         linestyles='dashed', levels=levels_pdf2d)

        if L2_ERR:
            # Compute L2 error
            dist = np.sqrt( np.sum((T_samp - fapp_samp)**2., axis=1) )
            log_entry.append( np.sqrt( np.sum( dist**2. ) / float(N_samp) ) )

        log.append( log_entry )

    if PLOTTING and PLOT_SUPPORT:
        fig_map_1d.gca().legend()
        fig_kde_1d.gca().legend()
        plt.show(False)

    # # Decorate map plotting
    # # ax_map.set_ylim([f(X[0]), f(X[-1])])
    # ax_map.scatter(x, np.zeros(x.shape))
    # ax_map.set_xlabel('x')
    # ax_map.set_ylabel('$T(x)$')
    # # ax_map.set_ylabel('$ (F^{-1} \circ F_{ref})x $')
    # # ax_map.set_ylim([np.min(exact_tm),np.max(exact_tm)])
    # ax_map.legend(loc='best')
    # ax_map.grid(True)

    # # fig_map.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-Approximation.pdf' % \
    # #                 (title, tit_type, tit_reg, tit_intest))
    # # fig_map.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-Approximation.eps' % \
    # #                 (title, tit_type, tit_reg, tit_intest))

    # # Decorate kde plotting
    # ax_kde.set_xlabel('x')
    # ax_kde.set_ylabel('PDF')
    # ax_kde.legend(loc='best')

    # # fig_kde.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-DistributionApproximation.pdf' % \
    # #                 (title, tit_type, tit_reg, tit_intest))
    # # fig_kde.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-DistributionApproximation.eps' % \
    # #                 (title, tit_type, tit_reg, tit_intest))

    # plt.show(False)

    print(" ")
    print(tabulate(log, headers=log_header))
    print(" ")

    print(tabulate(log, headers=log_header, tablefmt="latex", floatfmt=".2e"))

finally:
    if mpi_pool is not None:
        mpi_pool.stop()


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from . import BODDistributions
from .BODDistributions import *

__all__ = []
__all__ += BODDistributions.__all__


# This is the setup file for the amazing Xun module

from setuptools import setup
from setuptools.extension import Extension

if __name__ == "__main__":
    ext_modules = [ Extension('xun', 
                              ['src/xun.cpp'],
                              include_dirs = [],
                              # Extra symbols for compiler
                              extra_compile_args = ['-g', '-std=c++11'],
                              # Extra symbols for the linker
                              extra_link_args = ['-g']
                          )
                ]
    
    setup(
        name='xun',
        version = "1.0.0",
        license = "COPYING.LESSER",
        description = "The amazing Xun module",
        long_description=open("README.rst").read(),
        url="",
        author = "Xun",
        author_email = "xun@sandia",
        ext_modules = ext_modules,
        include_dirs=[],
        install_requires=['numpy'],
        zip_safe = False
    )




#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from TransportMaps.Distributions import Distribution

__all__ = ['BayesPosteriorDistribution']

class BayesPosteriorDistribution(Distribution):
    r""" Given a log-likelihood and a prior, assemble the posterior density

    Given the log-likelihood :math:`\log\pi({\bf y}\vert{\bf x})` and the
    prior density :math:`\pi({\bf x})`, assemble the Bayes' posterior density

    .. math::

      \pi({\bf x}\vert {\bf y}) \propto \pi({\bf y}\vert{\bf x}) \pi({\bf x})

    Args:
      logL (:class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`):
        log-likelihood :math:`\log\pi({\bf y}\vert{\bf x})`
      prior (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        prior density :math:`\pi({\bf x})`
    """
    def __init__(self, logL, prior):
        super(BayesPosteriorDistribution, self).__init__(prior.dim)
        self.prior = prior
        self.logL = logL

    @property
    def observations(self):
        return self.logL.y

    def log_pdf(self, x, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`\log\pi`
            at the ``x`` points.
        """
        return self.logL.evaluate(x, params, idxs_slice) \
            + self.prior.log_pdf(x, params, idxs_slice)

    def grad_x_log_pdf(self, x, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\nabla_{\bf x} \log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- values of
            :math:`\nabla_{\bf x}\log\pi` at the ``x`` points.
        """
        return self.logL.grad_x(x, params, idxs_slice) \
            + self.prior.grad_x_log_pdf(x, params, idxs_slice)

    def tuple_grad_x_log_pdf(self, x, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\left(\log \pi({\bf x}\vert{\bf y}), \nabla_{\bf x} \log \pi({\bf x}\vert{\bf y})\right)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`tuple`) --
            :math:`\left(\log \pi({\bf x}\vert{\bf y}), \nabla_{\bf x} \log \pi({\bf x}\vert{\bf y})\right)`
        """
        ll, gxll = self.logL.tuple_grad_x(x, params, idxs_slice)
        lpr, gxlpr = self.prior.tuple_grad_x_log_pdf(x, params, idxs_slice)
        ev = ll + lpr
        gx = gxll + gxlpr
        return (ev, gx)

    def hess_x_log_pdf(self, x, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\nabla^2_{\bf x} \log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) -- values of
            :math:`\nabla^2_{\bf x}\log\pi` at the ``x`` points.
        """
        return self.logL.hess_x(x, params, idxs_slice) \
            + self.prior.hess_x_log_pdf(x, params, idxs_slice)

            


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys
import os
import os.path
import dill
import numpy as np

from . import AvailableOptions as AO
from .ScriptBase import Script

from TransportMaps.External import H5PY_SUPPORT
import TransportMaps.Diagnostics as DIAG
import TransportMaps.Samplers as SAMP
import TransportMaps.Distributions as DIST

if H5PY_SUPPORT:
    import h5py

__all__ = ['Postprocess']

class Postprocess(Script):

    def usage(self):
        usage_str = """
Usage: tmap-postprocess [-h -v -I]
  --input=INPUT --output=OUTPUT
  [--store-fig-dir=DIR --store-fig-fmats=FMATS
   --extra-tit=TITLE --no-plotting 
   --aligned-conditionals=DIST
     --alc-n-points-x-ax=N --alc-n-tri-plots=N
     --alc-anchor=LIST --alc-range=LIST
   --random-conditionals=DIST
     --rndc-n-points-x-ax=N --rndc-n-plots-x-ax=N
     --rndc-anchor=LIST --rndc-range=LIST
   --var-diag=DIST
     --var-diag-qtype=QTYPE --var-diag-qnum=QNUM
   --aligned-marginals=DIST
     --alm-n-points=N --alm-n-tri-plots=N
   --quadrature=DIST
     --quadrature-qtype=QTYPE
     --quadrature-qnum=QNUM
   --importance-samples=NSAMP
   --mcmc=ALG
     --mcmc-samples=NSAMP
     --mcmc-burnin=BURNIN
     --mcmc-skip=SKIP
     --mcmc-no-ess --mcmc-ess-skip=SKIP
     --mcmc-ess-q=QUANTILE
     --mcmc-ess-corr-plot
     --mcmc-ess-corr-plot-lag=LAG
     --mcmc-ess-hist-plot
     --mcmc-hmc-eps=EPS
     --mcmc-hmc-nsteps=NSTEPS
   --chunk-size=SIZE --log=LOG --nprocs=NPROCS]
"""
        print(usage_str)

    def description(self):
        docs_distributions_str = \
            AO.print_avail_options(AO.AVAIL_DISTRIBUTIONS,
                                      '                        ', False)
        docs_mcmc_str = \
            '  --mcmc=ALG              algorithm to be used to generate Markov Chain.\n' + \
            AO.print_avail_options(AO.AVAIL_MCMC_ALGORITHMS,
                                      '                          ')
        docs_log_str = \
            '  --log=LOG               log level (default=30). Uses package logging.\n' + \
            AO.print_avail_options(AO.AVAIL_LOGGING,'                          ')
        docs_str = """DESCRIPTION
Given a file (--input) storing the transport map pushing forward a base distribution
to a target distribution, provides a number of diagnositic routines.
All files involved are stored and loaded using the python package dill and
an extra file OUTPUT.hdf5 is created to store big datasets in the hdf5 format.
In the following default values are shown in brackets.

OPTIONS - input/output:
  --input=INPUT         path to the file containing the target distribution,
                          the base distribution and the transport map pushing forward
                          the base to the target.
  --output=OUTPUT       path to the file storing all postprocess data.
                          The additional file OUTPUT.hdf5 will be used to store
                          the more memory consuming data.
  --store-fig-dir=DIR   path to the directory where to store the figures.
  --store-fig-fmats=FMATS  figure formats - see matplotlib for supported formats (svg)
  --extra-tit=TITLE     additional title for the figures' file names.
  --no-plotting         do not plot figures, but only store their data.
                          (requires --output or --store-fig-dir)
OPTIONS - Diagnostics:
  --aligned-conditionals=DIST  plot aligned slices of the selected DIST:
""" + docs_distributions_str + """                        Optional arguments:
    --alc-n-points-x-ax=N  number of discretization points per axis (40)
    --alc-n-tri-plots=N    number of subplots (0)
    --alc-anchor=LIST      list of floats "f1,f2,f3..." for the anchor point (0)
    --alc-range=LIST       list of two floats "f1,f2" for the range (-5,5)
  --random-conditionals=DIST   plot randomly chosen slices of the selected DIST:
""" + docs_distributions_str + """                        Optional arguments:
    --rndc-n-points-x-ax=N   number of discretization points per axis (40)
    --rndc-n-plots-x-ax=N    number of subplots (0)
    --rndc-anchor=LIST       list of floats "f1,f2,f3..." for the anchor point (0)
    --rndc-range=LIST        list of two floats "f1,f2" for the range (-5,5)
  --var-diag=DIST       compute variance diagostic using the sampling DIST:
""" + docs_distributions_str + """                        Optional arguments:
    --var-diag-qtype=QTYPE  quadrature type to be used (0)
    --var-diag-qnum=QNUM    level of the quadrature (1000)
OPTIONS - Sampling:
  --aligned-marginals=DIST  plot aligned marginals of the selected DIST:
""" + docs_distributions_str + """                        Optional arguments:
    --alm-n-points=N         number of samples to be used for the kernel density estimation
    --alm-n-tri-plots=N      number of subplots (0)
  --quadrature=DIST     generate quadrature for the selected DIST:
""" + docs_distributions_str + """                        Optional arguments:
    --quadrature-qtype=QTYPE  generate quadrature of type QTYPE (0)
    --quadrature-qnum=QNUM  level of the quadrature (int or list)
  --importance-samples=NSAMP  number of importance samples and weights for the approximation
                        of estimators with respect to the target distribution
""" + docs_mcmc_str + \
"""    --mcmc-samples=NSAMP  length of the chain with invariant distribution the
                          target distribution using Metropolis-Hastings with independent
                          proposals
    --mcmc-burnin=BURNIN   number of samples to be considered as burn-in
    --mcmc-skip=SKIP      number of sample to be skipped (>=0) in storage (a NSAMP*SKIP chain is subsampled)
    --mcmc-no-ess         turn off the computation of the effective sample size
    --mcmc-ess-skip=SKIP  number of samples to be skipped in the effective sample size estimation
    --mcmc-ess-q=QUANTILE  quantile used for the estimation of the sample size (0.99).
                             This is estimated over the worst decaying autocorrelation rate.
    --mcmc-ess-corr-plot   whether to plot the auto correlations
    --mcmc-ess-corr-plot-lag=LAG  maximum lag to be plotted (100)
    --mcmc-ess-hist-plot   whether to plot a summary of the sample size by dimension
    --mcmc-hmc-eps=EPS     epsilon value in Hamiltonian Monte Carlo
    --mcmc-hmc-nsteps=NSTEPS number of steps per sample
OPTIONS - Computation:
  --chunk-size=SIZE       chunk size to be used in the storage of data
""" + docs_log_str + """  --nprocs=NPROCS         number of processors to be used (default=1)
OPTIONS - other:
  -v                      verbose output (not affecting --log)
  -I                      enter interactive mode after finishing
  -h                      print this help
"""
        print(docs_str)

    def store_figure(self, fig, fname):
        for fmat in STORE_FIG_FMATS:
            fig.savefig(fname+'.'+fmat, format=fmat, bbox_inches='tight');
        
    @property
    def long_options(self):
        return super(Postprocess, self).long_options + \
            [
                # I/O
                "store-fig-dir=", "store-fig-fmats=",
                "extra-tit=", "no-plotting",
                # Aligned conditionals
                "aligned-conditionals=",
                "alc-n-points-x-ax=", "alc-n-tri-plots=", "alc-anchor=", "alc-range=",
                # Random conditionals
                "random-conditionals=",
                "rndc-n-points-x-ax=", "rndc-anchor=", "rndc-range=", "rndc-n-plots-x-ax=",
                # Aligned marginals
                "aligned-marginals=",
                "alm-n-points=", "alm-n-tri-plots=",
                # Variance diagnostic
                "var-diag=", "var-diag-qtype=", "var-diag-qnum=",
                # Quadrature
                "quadrature=", "quadrature-qtype=", "quadrature-qnum=",
                # Importance sampling
                "importance-samples=",
                # Markov Chain Monte Carlo
                "mcmc=",
                "mcmc-samples=", "mcmc-burnin=", "mcmc-skip=",
                "mcmc-hmc-eps=", "mcmc-hmc-nsteps=",
                "mcmc-no-ess", "mcmc-ess-skip=",
                "mcmc-ess-q=", "mcmc-ess-xcorr",
                "mcmc-ess-corr-plot", "mcmc-ess-corr-plot-lag=",
                "mcmc-ess-hist-plot",
                # Computation
                "chunk-size="
            ]

    def load_opts(self, opts):
        super(Postprocess, self).load_opts(opts)
        for opt, arg in opts:
            # I/O
            if opt == "--store-fig-dir":
                self.STORE_FIG_DIR = arg
            elif opt == "--store-fig-fmats":
                self.STORE_FIG_FMATS = arg.split(',')
            elif opt == "--extra-tit":
                self.EXTRA_TIT = "-" + arg
            elif opt == "--no-plotting":
                self.PLOTTING = False

            # Aligned conditionals
            elif opt == "--aligned-conditionals":
                self.ALIGNED_CONDITIONALS.append(arg)
                self.ALC_N_POINTS_X_AX.append( self.DFT_N_POINTS_X_AX )
                self.ALC_N_TRI_PLOTS.append( self.DFT_N_TRI_PLOTS )
                self.ALC_ANCHOR.append( self.DFT_ANCHOR )
                self.ALC_RANGE.append( self.DFT_RANGE )
            # Options
            elif opt == "--alc-n-points-x-ax":
                self.ALC_N_POINTS_X_AX[len(self.ALIGNED_CONDITIONALS)-1] = int(arg)
            elif opt == "--alc-n-tri-plots":
                self.ALC_N_TRI_PLOTS[len(self.ALIGNED_CONDITIONALS)-1] = list(range(int(arg)))
            elif opt == "--alc-anchor":
                self.ALC_ANCHOR[len(self.ALIGNED_CONDITIONALS)-1] = [float(s) for s in arg.split(',')]
            elif opt == "--alc-range":
                self.ALC_RANGE[len(self.ALIGNED_CONDITIONALS)-1] = [float(s) for s in arg.split(',')]

            # Random conditionals
            elif opt == "--random-conditionals":
                self.RANDOM_CONDITIONALS.append(arg)
                self.RNDC_N_POINTS_X_AX.append( self.DFT_N_POINTS_X_AX )
                self.RNDC_N_PLOTS_X_AX.append( self.DFT_N_PLOTS_X_AX )
                self.RNDC_ANCHOR.append( self.DFT_ANCHOR )
                self.RNDC_RANGE.append( self.DFT_RANGE )
            # Options
            elif opt == "--rndc-n-points-x-ax":
                self.RNDC_N_POINTS_X_AX[len(self.RANDOM_CONDITIONALS)-1] = int(arg)
            elif opt == "--rndc-n-plots-x-ax":
                self.RNDC_N_PLOTS_X_AX[len(self.RANDOM_CONDITIONALS)-1] = int(arg)
            elif opt == "--rndc-anchor":
                self.RNDC_ANCHOR[len(self.RANDOM_CONDITIONALS)-1] = [float(s) for s in arg.split(',')]
            elif opt == "--rndc-range":
                self.RNDC_RANGE[len(self.RANDOM_CONDITIONALS)-1] = [float(s) for s in arg.split(',')]

            # Aligned marginals
            elif opt == "--aligned-marginals":
                self.ALIGNED_MARGINALS.append(arg)
                self.ALM_N_POINTS.append(self.DFT_N_POINTS)
                self.ALM_N_TRI_PLOTS.append(self.DFT_N_TRI_PLOTS)
            # Options
            elif opt == "--alm-n-points":
                self.ALM_N_POINTS[len(self.ALIGNED_MARGINALS)-1] = int(arg)
            elif opt == "--alm-n-tri-plots":
                self.ALM_N_TRI_PLOTS[len(self.ALIGNED_MARGINALS)-1] = int(arg)

            # Variance diagnostic
            elif opt == "--var-diag":
                self.VAR_DIAG.append(arg)
                self.VD_QTYPE.append(self.DFT_VD_QTYPE)
                self.VD_QNUM.append(self.DFT_VD_QNUM)
            elif opt == "--var-diag-qtype":
                self.VD_QTYPE[len(self.VAR_DIAG)-1] = int(arg)
            elif opt == "--var-diag-qnum":
                self.VD_QNUM[len(self.VAR_DIAG)-1] = [int(q) for q in arg.split(',')]

            # Quadrature
            elif opt == "--quadrature":
                self.QUADRATURE.append( arg )
                self.QUAD_QTYPE.append( None )
                self.QUAD_QNUM.append( None )
            elif opt == "--quadrature-qtype":
                self.QUAD_QTYPE[len(self.QUADRATURE)-1] = int(arg)
            elif opt == "--quadrature-qnum":
                self.QUAD_QNUM[len(self.QUADRATURE)-1] = [int(q) for q in arg.split(',')]

            # Importance sampling
            elif opt == "--importance-samples":
                self.IMP_SAMPLES = int(arg)

            # Metropolis Hastings
            elif opt == "--mcmc":
                self.MCMC_ALG = arg
            elif opt == "--mcmc-samples":
                self.MCMC_SAMPLES = int(arg)
            elif opt == "--mcmc-burnin":
                self.MCMC_BURNIN = int(arg)
            elif opt == "--mcmc-skip":
                self.MCMC_SKIP = max(int(arg), 0)
            elif opt == "--mcmc-hmc-eps":
                self.MCMC_HMC_EPS = float(arg)
            elif opt == "--mcmc-hmc-nsteps":
                self.MCMC_HMC_NSTEPS = int(arg)
            elif opt == "--mcmc-no-ess":
                self.MCMC_DO_ESS = False
            elif opt == "--mcmc-ess-skip":
                self.MCMC_ESS_SKIP = int(arg)
                if self.MCMC_ESS_SKIP < 1:
                    raise ValueError("SKIP must be > 0 in --mcmc-ess-skip=SKIP")
            elif opt == "--mcmc-ess-q":
                self.MCMC_ESS_Q = float(arg)
            elif opt == "--mcmc-ess-corr-plot":
                self.MCMC_ESS_CORR_PLOT = True
            elif opt == "--mcmc-ess-corr-plot-lag":
                self.MCMC_ESS_CORR_PLOT_LAG = int(arg)
            elif opt == "--mcmc-ess-hist-plot":
                self.MCMC_ESS_HIST_PLOT = True

            # Computation
            elif opt == "--chunk-size":
                self.CHUNK_SIZE = int(arg)

    def __init__(self, argv):
        # I/O
        self.STORE_FIG_DIR = None
        self.STORE_FIG_FMATS = ['svg']
        self.EXTRA_TIT = ''
        self.PLOTTING = True
        # Aligned conditionals
        self.ALIGNED_CONDITIONALS = []
        self.ALC_N_POINTS_X_AX = []
        self.ALC_N_TRI_PLOTS = []
        self.ALC_ANCHOR = []
        self.ALC_RANGE = []
        # Random conditionals
        self.RANDOM_CONDITIONALS = []
        self.RNDC_N_POINTS_X_AX = []
        self.RNDC_N_PLOTS_X_AX = []
        self.RNDC_ANCHOR = []
        self.RNDC_RANGE = []
        # Aligned marginals
        self.ALIGNED_MARGINALS = []
        self.ALM_N_POINTS = []
        self.ALM_N_TRI_PLOTS = []
        # Default plotting options
        self.DFT_N_POINTS = 1000
        self.DFT_N_POINTS_X_AX = 40
        self.DFT_N_TRI_PLOTS = 0
        self.DFT_ANCHOR = None
        self.DFT_RANGE = [-5.,5.]
        self.DFT_N_PLOTS_X_AX = 6
        # Variance diagnostic
        self.VAR_DIAG = []
        self.VD_QTYPE = []
        self.VD_QNUM  = []
        # Defaults for variance diagnostic
        self.DFT_VD_QTYPE = 0
        self.DFT_VD_QNUM = 1000
        # Samples
        self.QUADRATURE = []
        self.QUAD_QTYPE = []
        self.QUAD_QNUM = []
        # Importance samples
        self.IMP_SAMPLES = None
        # MCMC samples
        self.MCMC_ALG = None
        self.MCMC_SAMPLES = None
        self.MCMC_BURNIN = 0
        self.MCMC_SKIP = 0
        self.MCMC_DO_ESS = True
        self.MCMC_ESS_SKIP = 1
        self.MCMC_ESS_Q = 0.99
        self.MCMC_ESS_CORR_PLOT = False
        self.MCMC_ESS_CORR_PLOT_LAG = 100
        self.MCMC_ESS_HIST_PLOT = False
        self.MCMC_HMC_EPS = 0.2
        self.MCMC_HMC_NSTEPS = 1
        # hdf5 options
        self.CHUNK_SIZE = 10000

        super(Postprocess, self).__init__(argv)

        if not self.PLOTTING and self.STORE_FIG_DIR is None and self.OUTPUT is None:
            self.usage()
            self.tstamp_print(
                "ERROR: Neither --output nor --store-fig-dir were " + \
                "specified, while --no-plotting is active. " + \
                "This would result on no data shown or stored.")
            sys.exit(3)

        # Prepare storage of figures
        if self.STORE_FIG_DIR is not None:
            tit_no_path = str.split(DATA,"/")[-1]
            self.TITLE = '.'.join(str.split(tit_no_path,".")[:-1])

    def load(self):
        self.h5_file = None
        
        # Load data
        with open(self.INPUT, 'rb') as in_stream:
            self.stg = dill.load(in_stream)

        # Restore data
        self.base_distribution = self.stg.base_distribution
        self.target_distribution = self.stg.target_distribution
        self.tmap = self.stg.tmap
        self.approx_base_distribution = self.stg.approx_base_distribution
        self.approx_target_distribution = self.stg.approx_target_distribution
        self.dim = self.base_distribution.dim

        # Load output (dill file) if any
        if not os.path.exists(self.OUTPUT):
            self.postproc_data = {}
            with open(self.OUTPUT, 'wb') as out_stream:
                dill.dump(self.postproc_data, out_stream)
        with open(self.OUTPUT, 'rb') as in_stream:
            self.postproc_data = dill.load(in_stream)
        # Load output (hdf5 file) if any
        self.h5_file = h5py.File(self.OUTPUT + '.hdf5', 'a')
        self.h5_root = self.h5_file

    def aligned_conditionals(self, mpi_pool):
        for aligned, n_tri_plots, n_points_x_ax, anchor, rng in \
            zip(self.ALIGNED_CONDITIONALS, self.ALC_N_TRI_PLOTS,
                self.ALC_N_POINTS_X_AX, self.ALC_ANCHOR, self.ALC_RANGE):
            self.filter_tstamp_print("[Start] Aligned conditionals " + aligned)
            if aligned == 'exact-target':
                d = self.target_distribution
            elif aligned == 'approx-target':
                d = self.approx_target_distribution
            elif aligned == 'exact-base':
                d = self.base_distribution
            elif aligned == 'approx-base':
                d = self.approx_base_distribution
            else:
                self.full_usage()
                self.tstamp_print("ERROR: DIST %s not recognized." % aligned)
                sys.exit(3)
            DATA_FIELD = 'aligned-' + aligned
            data = self.postproc_data.get(DATA_FIELD, None)
            if data is None:
                data = DIAG.computeAlignedConditionals(
                    d, dimensions_vec=n_tri_plots,
                    numPointsXax = n_points_x_ax,
                    pointEval=anchor, range_vec=rng,
                    mpi_pool=mpi_pool)
                self.postproc_data[DATA_FIELD] = data
                if self.OUTPUT is not None:
                    self.safe_store(self.postproc_data, self.OUTPUT)
            if self.PLOTTING:
                fig = DIAG.plotAlignedConditionals(
                    data=data, show_flag=(self.STORE_FIG_DIR is None));
                if self.STORE_FIG_DIR is not None:
                    self.store_figure(fig, self.STORE_FIG_DIR+'/'+self.TITLE+ \
                                 '-aligned-conditionals-'+ aligned +\
                                 self.EXTRA_TIT)
            self.filter_tstamp_print("[Stop]  Aligned conditionals " + aligned)

    def random_conditionals(self, mpi_pool):
        for random, n_plots_x_ax, n_points_x_ax, anchor, rng in \
            zip(self.RANDOM_CONDITIONALS, self.RNDC_N_PLOTS_X_AX,
                self.RNDC_N_POINTS_X_AX,
                self.RNDC_ANCHOR, self.RNDC_RANGE):
            self.filter_tstamp_print("[Start] Random conditionals " + random)
            if random == 'exact-target':
                d = self.target_distribution
            elif random == 'approx-target':
                d = self.approx_target_distribution
            elif random == 'exact-base':
                d = self.base_distribution
            elif random == 'approx-base':
                d = self.approx_base_distribution
            else:
                self.full_usage()
                self.tstamp_print("ERROR: DIST %s not recognized." % random)
                sys.exit(3)
            DATA_FIELD = 'random-' + random
            data = self.postproc_data.get(DATA_FIELD, None)
            if data is None:
                data = DIAG.computeRandomConditionals(
                    d, num_conditionalsXax=n_plots_x_ax,
                    numPointsXax=n_points_x_ax,
                    pointEval=anchor, range_vec=rng,
                    mpi_pool=mpi_pool)
                self.postproc_data[DATA_FIELD] = data
                if self.OUTPUT is not None:
                    self.safe_store(self.postproc_data, self.OUTPUT)
            if self.PLOTTING:
                fig = DIAG.plotRandomConditionals(
                    data=data, show_flag=(self.STORE_FIG_DIR is None))
                if self.STORE_FIG_DIR is not None:
                    self.store_figure(fig, self.STORE_FIG_DIR+'/'+\
                                      self.TITLE+'-random-conditionals-' + random + \
                                      self.EXTRA_TIT)
            self.filter_tstamp_print("[Stop]  Random conditionals " + random)

    def variance_diagnostic(self, mpi_pool):
        for dstr, qtype, qnum in zip(self.VAR_DIAG, self.VD_QTYPE, self.VD_QNUM):
            self.filter_tstamp_print("[Start] Variance diagnostic " + dstr)
            if dstr == 'exact-target':
                d1 = self.target_distribution
                d2 = self.approx_target_distribution
            elif dstr == 'approx-target':
                d1 = self.approx_target_distribution
                d2 = self.target_distribution
            elif dstr == 'exact-base':
                d1 = self.base_distribution
                d2 = self.approx_base_distribution
            elif dstr == 'approx-base':
                d1 = self.approx_base_distribution
                d2 = self.base_distribution
            else:
                self.full_usage()
                self.tstamp_print("ERROR: DIST %s not recognized." % dstr)
                sys.exit(3)
            # Load values if any
            GRP_NAME = "vals_var_diag/" + dstr
            if GRP_NAME not in self.h5_root:
                self.h5_root.create_group(GRP_NAME)
            grp = self.h5_root[GRP_NAME]
            QTYPE_NAME = str(qtype)
            if QTYPE_NAME not in grp:
                grp.create_group(QTYPE_NAME)
            qtype_grp = grp[QTYPE_NAME]
            V1_NAME = 'vals_d1'
            V2_NAME = 'vals_d2'
            if qtype == 0: # Monte-Carlo
                if V1_NAME not in qtype_grp:
                    qtype_grp.create_dataset(
                        V1_NAME, (0,), maxshape=(None,), dtype='d', chunks=(self.CHUNK_SIZE,))
                if V2_NAME not in qtype_grp:
                    qtype_grp.create_dataset(
                        V2_NAME, (0,), maxshape=(None,), dtype='d', chunks=(self.CHUNK_SIZE,))
                loaded_vals_d1 = qtype_grp[V1_NAME]
                loaded_vals_d2 = qtype_grp[V2_NAME]
                if len(loaded_vals_d1) > 0 and qnum[0] <= len(loaded_vals_d1):
                    # Subselect already available data
                    vals_d1 = np.array( loaded_vals_d1[:qnum[0]] )
                    vals_d2 = np.array( loaded_vals_d2[:qnum[0]] )
                else:
                    old_len = len(loaded_vals_d1)
                    # Sample new points and evaluate
                    n_new_samps = qnum[0] - len(loaded_vals_d1)
                    x = d1.rvs(n_new_samps)
                    new_vals_d1, new_vals_d2 = DIAG.compute_vals_variance_approx_kl(
                        d1, d2, x=x, mpi_pool_tuple=(None, mpi_pool))
                    loaded_vals_d1.resize(qnum[0], axis=0)
                    loaded_vals_d2.resize(qnum[0], axis=0)
                    loaded_vals_d1[old_len:] = new_vals_d1
                    loaded_vals_d2[old_len:] = new_vals_d2
                    vals_d1 = np.array( loaded_vals_d1 )
                    vals_d2 = np.array( loaded_vals_d2 )
                w = np.ones(qnum[0])/float(qnum[0])
            elif qtype == 3: # Gauss quadrature
                QNUM_NAME = str(qnum)
                W_NAME = 'w'
                if QNUM_NAME not in qtype_grp:
                    qtype_grp.create_group(QNUM_NAME)
                    qnum_grp = qtype_grp[QNUM_NAME]
                    (x, w) = d1.quadrature(qtype, qnum, mpi_pool=mpi_pool)
                    vals_d1, vals_d2 = DIAG.compute_vals_variance_approx_kl(
                        d1, d2, x=x, mpi_pool_tuple=(None, mpi_pool))
                    csize = min(self.CHUNK_SIZE, w.shape[0])
                    qnum_grp.create_dataset(V1_NAME, data=vals_d1, chunks=(csize,))
                    qnum_grp.create_dataset(V2_NAME, data=vals_d2, chunks=(csize,))
                    qnum_grp.create_dataset(W_NAME, data=w, chunks=(csize,))
                else:
                    qnum_grp[QNUM_NAME]
                    vals_d1 = np.array( qnum_grp[V1_NAME] )
                    vals_d2 = np.array( qnum_grp[V2_NAME] )
                    w = np.array( qnum_grp[W_NAME] )
            var_diag_tm = DIAG.variance_approx_kl(d1, d2,
                                                  vals_d1=vals_d1, vals_d2=vals_d2, w=w)
            self.filter_tstamp_print("[Stop]  Variance diagnostic %s: %e" % (dstr, var_diag_tm))

    def aligned_marginals(self, mpi_pool):
        for dstr, n_points, n_tri_plots in zip(
                self.ALIGNED_MARGINALS, self.ALM_N_POINTS, self.ALM_N_TRI_PLOTS):
            self.filter_tstamp_print("[Start] Aligned marginals %s " % dstr + \
                           "- Sample generation")
            if dstr == 'exact-target':
                d = self.target_distribution
            elif dstr == 'approx-target':
                d = self.approx_target_distribution
            elif dstr == 'exact-base':
                d = self.base_distribution
            elif dstr == 'approx-base':
                d = self.approx_base_distribution
            else:
                self.full_usage()
                self.tstamp_print("ERROR: DIST %s not recognized." % dstr)
                sys.exit(3)
            # Load values if any
            Q_GRP_NAME = "quadrature"
            if Q_GRP_NAME not in self.h5_root:
                self.h5_root.create_group(Q_GRP_NAME)
            qgrp = self.h5_root[Q_GRP_NAME]
            D_GRP_NAME = dstr
            if D_GRP_NAME not in qgrp:
                qgrp.create_group(D_GRP_NAME)
            dgrp = qgrp[D_GRP_NAME]
            DSET_NAME = '0'
            if DSET_NAME not in dgrp:
                dgrp.create_dataset(
                    DSET_NAME, (0,self.dim), maxshape=(None,self.dim), dtype='d',
                    chunks=(self.CHUNK_SIZE,1))
            loaded_samp = dgrp[DSET_NAME]
            if n_points > loaded_samp.shape[0]:
                nold = loaded_samp.shape[0]
                new_nsamp = n_points - nold
                x = d.rvs(new_nsamp, mpi_pool=mpi_pool)
                loaded_samp.resize(n_points, axis=0)
                loaded_samp[nold:,:] = x
            self.filter_tstamp_print("        Aligned marginals %s " % dstr + \
                                "- Plotting")
            if self.PLOTTING:
                fig = DIAG.plotAlignedMarginals(
                    loaded_samp[:n_points,:], n_tri_plots,
                    show_flag=(self.STORE_FIG_DIR is None))
                if self.STORE_FIG_DIR is not None:
                    self.store_figure(fig, self.STORE_FIG_DIR+'/'+\
                                 self.TITLE+'-aligned-marginals-'+ dstr +\
                                 self.EXTRA_TIT)
            self.filter_tstamp_print("[Stop]  Aligned marginals %s" % dstr)

    def quadratures(self, mpi_pool):
        for dstr, qtype, qnum in zip(self.QUADRATURE, self.QUAD_QTYPE, self.QUAD_QNUM):
            self.filter_tstamp_print("[Start] Quadrature " + str(qtype))
            if dstr == 'exact-target':
                d = self.target_distribution
            elif dstr == 'approx-target':
                d = self.approx_target_distribution
            elif dstr == 'exact-base':
                d = self.base_distribution
            elif dstr == 'approx-base':
                d = self.approx_base_distribution
            else:
                self.full_usage()
                self.tstamp_print("ERROR: DIST %s not recognized." % dstr)
                sys.exit(3)
            # Load values if any
            GRP_NAME = "quadrature"
            if GRP_NAME not in self.h5_root:
                self.h5_root.create_group(GRP_NAME)
            qgrp = self.h5_root[GRP_NAME]
            D_GRP_NAME = dstr
            if D_GRP_NAME not in qgrp:
                qgrp.create_group(D_GRP_NAME)
            dgrp = qgrp[D_GRP_NAME]
            if qtype == 0: # Monte-Carlo
                DSET_NAME = str(qtype)
                if DSET_NAME not in dgrp:
                    dgrp.create_dataset(
                        DSET_NAME, (0,self.dim), maxshape=(None,self.dim), dtype='d',
                        chunks=(self.CHUNK_SIZE,1))
                loaded_samp = dgrp[DSET_NAME]
                if qnum[0] > loaded_samp.shape[0]:
                    nold = loaded_samp.shape[0]
                    new_nsamp = qnum[0] - nold
                    x = d.rvs(new_nsamp, mpi_pool=mpi_pool)
                    loaded_samp.resize(qnum[0], axis=0)
                    loaded_samp[nold:,:] = x
            elif qtype == 3: # Gauss quadrature
                QTYPE_NAME = str(qtype)
                X_NAME = 'x'
                W_NAME = 'w'
                if QTYPE_NAME not in dgrp:
                    dgrp.create_group(QTYPE_NAME)
                qtp_grp = dgrp[QTYPE_NAME]
                QNUM_NAME = str(qnum)
                if QNUM_NAME not in qtp_grp:
                    qtp_grp.create_group(QNUM_NAME)
                    qngrp = qtp_grp[QNUM_NAME]
                    (x, w) = d.quadrature(qtype, qnum, mpi_pool=mpi_pool)
                    qngrp.create_dataset(X_NAME, data=x, chunks=(self.CHUNK_SIZE,1))
                    qngrp.create_dataset(W_NAME, data=w, chunks=(self.CHUNK_SIZE,))
            self.filter_tstamp_print("[Stop]  Quadrature")

    def importance_sampling(self, mpi_pool):
        if self.IMP_SAMPLES is not None:
            self.filter_tstamp_print("[Start] Importance sampling")
            # Load values if any
            GRP_NAME = "importance-samples"
            if GRP_NAME not in self.h5_root:
                self.h5_root.create_group(GRP_NAME)
            is_grp = self.h5_root[GRP_NAME]
            X_NAME = 'x'
            W_NAME = 'w'
            if X_NAME not in is_grp:
                is_grp.create_dataset(
                    X_NAME, (0,self.dim), maxshape=(None,self.dim),
                    dtype='d', chunks=(self.CHUNK_SIZE,1))
                is_grp.create_dataset(
                    W_NAME, (0,), maxshape=(None,),
                    dtype='d', chunks=(self.CHUNK_SIZE,))
            loaded_x = is_grp[X_NAME]
            loaded_w = is_grp[W_NAME]
            if self.IMP_SAMPLES > loaded_x.shape[0]:
                nold = loaded_x.shape[0]
                new_nsamp = self.IMP_SAMPLES - nold
                sampler = SAMP.ImportanceSampler( self.approx_base_distribution, self.base_distribution )
                (x, w) = sampler.rvs(new_nsamp, mpi_pool_tuple=(mpi_pool, None))
                x = self.approx_target_distribution.map_samples_base_to_target(
                    x, mpi_pool=mpi_pool)
                loaded_x.resize(self.IMP_SAMPLES, axis=0)
                loaded_x[nold:,:] = x
                loaded_w.resize(self.IMP_SAMPLES, axis=0)
                loaded_w[nold:] = w
                loaded_w /= np.sum(loaded_w)
            self.filter_tstamp_print("[Stop]  Importance sampling")

    def mcmc(self, mpi_pool):
        if self.MCMC_ALG is not None:
            self.filter_tstamp_print("[Start] Markov Chain Monte Carlo")
            if self.MCMC_ALG == 'mhind':
                self.filter_tstamp_print("        Metropolis-Hastings with Independent Proposals")
                GRP_NAME = "metropolis-independent-proposal-samples"
                sampler = SAMP.MetropolisHastingsIndependentProposalsSampler(
                    self.approx_base_distribution, self.base_distribution )
                # Load values if any
                if GRP_NAME not in self.h5_root:
                    self.h5_root.create_group(GRP_NAME)
                is_grp = self.h5_root[GRP_NAME]
            elif self.MCMC_ALG == 'hmc':
                self.filter_tstamp_print("        Hamiltonian Monte Carlo")
                GRP_NAME = "hamiltonian-monte-carlo-samples"
                if not isinstance(self.base_distribution, DIST.StandardNormalDistribution):
                    self.logger.warn("The HMC algorithm uses a Standard Normal distribution " +\
                                     "as default proposal")
                sampler = SAMP.HamiltonianMonteCarloSampler( self.approx_base_distribution )
                # Load values if any
                if GRP_NAME not in self.h5_root:
                    self.h5_root.create_group(GRP_NAME)
                is_grp = self.h5_root[GRP_NAME]
                if str(self.MCMC_HMC_EPS) not in is_grp:
                    is_grp.create_group(str(self.MCMC_HMC_EPS))
                is_grp = is_grp[str(self.MCMC_HMC_EPS)]
                if str(self.MCMC_HMC_NSTEPS) not in is_grp:
                    is_grp.create_group(str(self.MCMC_HMC_NSTEPS))
                is_grp = is_grp[str(self.MCMC_HMC_NSTEPS)]
            else:
                self.full_usage()
                self.tstamp_print("ERROR: ALG %s not recognized." % MCMC_ALG)
                sys.exit(3)
            # Create group for each skipping value
            SKIP_NAME = "skip-%d" % self.MCMC_SKIP
            if SKIP_NAME not in is_grp:
                is_grp.create_group(SKIP_NAME)
            is_grp = is_grp[SKIP_NAME]
            X_NAME = 'x'  # Samples in pushforward space
            S_NAME = 's'  # Samples in pullback space
            if X_NAME not in is_grp:
                is_grp.create_dataset(
                    X_NAME, (0,self.dim), maxshape=(None,self.dim),
                    dtype='d', chunks=(self.CHUNK_SIZE,1))
            if S_NAME not in is_grp:
                is_grp.create_dataset(
                    S_NAME, (0,self.dim), maxshape=(None,self.dim),
                    dtype='d', chunks=(self.CHUNK_SIZE,1))
            loaded_x = is_grp[X_NAME]
            loaded_s = is_grp[S_NAME]
            if self.MCMC_SAMPLES > loaded_x.shape[0]:
                nold = loaded_x.shape[0]
                new_nsamp = self.MCMC_SAMPLES - nold
                s0 = None
                if nold > 0:
                    s0 = loaded_s[-1,:]
                    self.filter_tstamp_print("        Restarting chain from stored data (length: %d)" % nold)
                if self.MCMC_ALG == 'mhind':
                    (s, _) = sampler.rvs(new_nsamp*(self.MCMC_SKIP+1), x0=s0,
                                         mpi_pool_tuple=(mpi_pool, None))
                elif self.MCMC_ALG == 'hmc':
                    (s, _) = sampler.rvs(
                        new_nsamp*(self.MCMC_SKIP+1), x0=s0,
                        epsilon=self.MCMC_HMC_EPS, n_steps=self.MCMC_HMC_NSTEPS)
                s = s[::(self.MCMC_SKIP+1),:]
                x = self.approx_target_distribution.map_samples_base_to_target(
                    s, mpi_pool=mpi_pool)
                loaded_x.resize(self.MCMC_SAMPLES, axis=0)
                loaded_s.resize(self.MCMC_SAMPLES, axis=0)
                loaded_x[nold:,:] = x
                loaded_s[nold:,:] = s
            # Compute effective sample size
            if self.MCMC_DO_ESS:
                self.filter_tstamp_print("        Estimating ESS")
                ess_list = []
                for d in range(loaded_x.shape[1]):
                    fig = None if not (self.PLOTTING and self.MCMC_ESS_CORR_PLOT) else plt.figure()
                    ess = SAMP.ess(loaded_x[self.MCMC_BURNIN:self.MCMC_SAMPLES:self.MCMC_ESS_SKIP,[d]],
                                   quantile=self.MCMC_ESS_Q,
                                   plotting=(self.PLOTTING and self.MCMC_ESS_CORR_PLOT),
                                   plot_lag=self.MCMC_ESS_CORR_PLOT_LAG, fig=fig)
                    self.filter_tstamp_print(
                        "        ESS dimension %d: %d/%d " % (
                            d,ess,(self.MCMC_SAMPLES-self.MCMC_BURNIN)//self.MCMC_ESS_SKIP) + \
                        "(%2.3f%%)" % (
                            ess/float((self.MCMC_SAMPLES-self.MCMC_BURNIN)//self.MCMC_ESS_SKIP)*100.)
                    )
                    ess_list.append( ess )
                    if self.PLOTTING and self.MCMC_ESS_CORR_PLOT:
                        if self.STORE_FIG_DIR is None:
                            plt.show(False)
                        else:
                            self.store_figure(fig, self.STORE_FIG_DIR+'/'+\
                                              self.TITLE+'-metropolis-ess-corr-d%d' %d + \
                                              self.EXTRA_TIT)
                if self.PLOTTING and self.MCMC_ESS_HIST_PLOT:
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    ax.plot(ess, drawstyle='steps-mid')
                    ax.set_ylabel("ESS")
                    ax.set_xlable("Dimension")
                    if self.STORE_FIG_DIR is None:
                        plt.show(False)
                    else:
                        self.store_figure(fig, self.STORE_FIG_DIR+'/'+\
                                          self.TITLE+'-metropolis-ess-hist' + \
                                          self.EXTRA_TIT)
                ess = min(ess_list) # We report the worst ess
                self.filter_tstamp_print(
                    "[Stop]  Markov Chain Monte Carlo " + \
                    "- ESS: %d/%d " % (ess,(self.MCMC_SAMPLES-self.MCMC_BURNIN)//self.MCMC_ESS_SKIP) + \
                    "(%2.3f%%)" % (ess/float((self.MCMC_SAMPLES-self.MCMC_BURNIN)//self.MCMC_ESS_SKIP)*100.) )

            else:
                self.filter_tstamp_print("[Stop]  Markov Chain Monte Carlo")

    def run(self, mpi_pool):
        self.aligned_conditionals(mpi_pool)
        self.random_conditionals(mpi_pool)
        self.variance_diagnostic(mpi_pool)
        self.aligned_marginals(mpi_pool)
        self.quadratures(mpi_pool)
        self.importance_sampling(mpi_pool)
        self.mcmc(mpi_pool)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import TransportMaps as TM

__all__ = ['Function']

class Function(TM.TMO):
    r""" Abstract class for parametric approximation :math:`f_{\bf a}:\mathbb{R}^d\rightarrow\mathbb{R}` of :math:`f:\mathbb{R}^d\rightarrow\mathbb{R}`.

    Args:
      dim (int): number of dimensions
    """
    def __init__(self, dim):
        super(Function, self).__init__()
        self.dim = dim

    def evaluate(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function evaluations
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_x(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf x} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def partial_xd(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def hess_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def partial2_xd(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`\partial^2_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial^2_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import numpy as np

import TransportMaps as TM

__all__ = ['Map', 'ParametricMap',
           'LinearMap', 'ConditionallyLinearMap']

nax = np.newaxis

class Map(TM.TMO):
    r""" Abstract map :math:`T:\mathbb{R}^{d_x}\rightarrow\mathbb{R}^{d_y}`

    Args:
      dim_in (int): input dimension :math:`d_x`
      dim_out (int): output dimension :math:`d_y`
    """
    def __init__(self, dim_in, dim_out):
        super(Map, self).__init__()
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.dim = None
        if self.dim_in == self.dim_out:
            self.dim = self.dim_in

    def __call__(self, x):
        r"""
        Calls :func:`evaluate`.
        """
        return self.evaluate( x )

    def evaluate(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate the map :math:`T` at the points :math:`{\bf x} \in \mathbb{R}^{m \times d_x}`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d_x`]): evaluation points
          precomp (:class:`dict<dict>`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]) -- transformed points

        Raises:
          NotImplementedError: to be implemented in sub-classes
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_x(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate the gradient :math:`\nabla_{\bf x}T` at the points :math:`{\bf x} \in \mathbb{R}^{m \times d_x}`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d_x`]): evaluation points
          precomp (:class:`dict<dict>`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y,d_y`]) -- transformed points

        Raises:
          NotImplementedError: to be implemented in sub-classes
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def hess_x(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate the Hessian :math:`\nabla^2_{\bf x}T` at the points :math:`{\bf x} \in \mathbb{R}^{m \times d_x}`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d_x`]): evaluation points
          precomp (:class:`dict<dict>`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y,d_y,d_y`]) -- transformed points

        Raises:
          NotImplementedError: to be implemented in sub-classes
        """
        raise NotImplementedError("To be implemented in sub-classes")

class ParametricMap(Map):
    r""" Abstract map :math:`T:\mathbb{R}^{d_a}\times\mathbb{R}^{d_x}\rightarrow\mathbb{R}^{d_y}`

    Args:
      dim_in (int): input dimension :math:`d_x`
      dim_out (int): output dimension :math:`d_y`
    """    
    @property
    def n_coeffs(self):
        r""" Returns the total number of coefficients.

        Returns:
          (:class:`int`) -- total number :math:`N` of
              coefficients characterizing the map.

        Raises:
          NotImplementedError: needs to be implemented in subclasses
        """
        raise NotImplementedError("Must be implemented in subclasses")

    def get_n_coeffs(self):
        TM.deprecate("ParametricMap.get_n_coeffs()", "1.0b3",
                     "Use property ParametricMap.n_coeffs instead")
        return self.n_coeffs
        
    @property
    def coeffs(self):
        r""" Returns the actual value of the coefficients.

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`N`]) -- coefficients.

        Raises:
          NotImplementedError: needs to be implemented in subclasses
        """
        raise NotImplementedError("Must be implemented in subclasses")

    def get_coeffs(self):
        TM.deprecate("ParametricMap.get_coeffs()", "1.0b3",
                     "Use property ParametricMap.coeffs instead")
        return self.coeffs
    
    @coeffs.setter
    def coeffs(self, coeffs):
        r""" Set the coefficients.

        Args:
           coeffs (:class:`ndarray<numpy.ndarray>` [:math:`N`]):
              coefficients for the various maps

        Raises:
          NotImplementedError: needs to be implemented in subclasses
        """
        raise NotImplementedError("Must be implemented in subclasses")

    def set_coeffs(self, coeffs):
        TM.deprecate("ParametricMap.set_coeffs(value)", "1.0b3",
                     "Use setter ParametricMap.coeffs = value instead.")
        self.coeffs = coeffs

    def grad_a(self, x, precomp=None, idxs_slice=slice(None)):
        r""" Compute :math:`\nabla_{\bf a} T({\bf x},{\bf a})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict<dict>`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          
        Returns:
           (:class:`ndarray<numpy.ndarray>`) -- gradient

        Raises:
          NotImplementedError: needs to be implemented in subclasses
        """
        raise NotImplementedError("Must be implemented in subclasses")

    def hess_a(self, x, precomp=None, idxs_slice=slice(None)):
        r""" Compute :math:`\nabla^2_{\bf a} T({\bf x},{\bf a})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict<dict>`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          
        Returns:
           (:class:`ndarray<numpy.ndarray>`) -- Hessian

        Raises:
          NotImplementedError: needs to be implemented in subclasses
        """
        raise NotImplementedError("Must be implemented in subclasses")
        
class LinearMap(Map):
    r""" Map :math:`T({\bf x}) = {\bf c} + {\bf T} {\bf x}`

    Args:
      c (:class:`ndarray<numpy.ndarray>` [:math:`d_y`]): constant part
      T (:class:`ndarray<numpy.ndarray>` [:math:`d_y,d_x`]): linear part (matrix)
    """
    def __init__(self, c, T):
        if c.shape[0] != T.shape[0]:
            raise ValueError("The dimensions of the constant and the " + \
                             "linear part must match")
        super(LinearMap, self).__init__(T.shape[1], T.shape[0])
        self._c = c
        self._T = T

    @property
    def c(self):
        return self._c

    @property
    def T(self):
        return self._T

    def evaluate(self, x, *args, **kwargs):
        return self.c + self.T.dot(x.T).T

    def grad_x(self, x, *args, **kwargs):
        return self.T[nax,:,:]

    def hess_x(self, x, *args, **kwargs):
        return np.zeros((1, self.dim_out, self.dim_in, self.dim_in))

class ConditionallyLinearMap(Map):
    r""" Map :math:`T:\mathbb{R}^{d_x}\times\mathbb{R}^{d_a}\rightarrow\mathbb{R}^{d_y}` defined by :math:`T({\bf x};{\bf a}) = {\bf c}({\bf a}) + {\bf T}({\bf a}) {\bf x}`

    Args:
      c (:class:`Map`): map :math:`{\bf c}:\mathbb{R}^{d_a}\rightarrow\mathbb{R}^{d_y}`
      T (:class:`Map`):
        map :math:`{\bf T}:\mathbb{R}^{d_a}\rightarrow\mathbb{R}^{d_y\times d_x}`
      coeffs (:class:`ndarray<numpy.ndarray>`): fixing the coefficients :math:`{\bf a}` defining
        :math:`{\bf c}({\bf a})` and :math:`{\bf T}({\bf a})`.
    """
    def __init__(self, c, T, coeffs=None):
        if c.dim_in != T.dim_in:
            raise ValueError("Input dimension mismatch between c and T")
        if T.dim_out % c.dim_out != 0:
            raise ValueError("Output dimension mismatch between c and T")
        self._n_coeffs = c.dim_in
        self._cMap = c
        self._TMap = T
        din = T.dim_out // c.dim_out
        dout = c.dim_out
        super(ConditionallyLinearMap,self).__init__(
            din + self.n_coeffs, dout)
        self._coeffs = None
        self.coeffs = coeffs
        
    @property
    def c(self):
        return self._c

    @property
    def T(self):
        return self._T
        
    @property
    def n_coeffs(self):
        return self._n_coeffs

    @property
    def dim_lin(self):
        return self.dim_in - self.n_coeffs

    @property
    def coeffs(self):
        r""" Returns the actual value of the coefficients.

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`N`]) -- coefficients.
        """
        return self._coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        r""" Set the coefficients.

        Args:
           coeffs (:class:`ndarray<numpy.ndarray>` [:math:`N`]):
              coefficients for the various maps
        """
        if coeffs is None:
            self._coeffs = None
        elif self._coeffs is None or np.any(self._coeffs != coeffs):
            self._c = self._cMap.evaluate(coeffs[nax,:])[0,:]
            self._T = self._TMap.evaluate(coeffs[nax,:])[0,:,:]
            try:
                self._grad_a_c = self._cMap.grad_x(coeffs[nax,:])[0,:,:]
                self._grad_a_T = self._TMap.grad_x(coeffs[nax,:])[0,:,:,:]
            except NotImplementedError:
                self._grad_a_c = None
                self._grad_a_T = None
            try:
                self._hess_a_c = self._cMap.hess_x(coeffs[nax,:])[0,:,:,:]
                self._hess_a_T = self._TMap.hess_x(coeffs[nax,:])[0,:,:,:,:]
            except NotImplementedError:
                self._hess_a_c = None
                self._hess_a_T = None
            self._coeffs = coeffs

    @property
    def grad_a_c(self):
        return self._grad_a_c

    @property
    def grad_a_T(self):
        return self._grad_a_T

    @property
    def hess_a_c(self):
        return self._hess_a_c

    @property
    def hess_a_T(self):
        return self._hess_a_T

    def evaluate(self, x, precomp=None, idxs_slice=slice(None)):
        r""" Evaluate the map :math:`T` at the points :math:`{\bf x}`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d_x`]): evaluation points
          precomp (:class:`dict<dict>`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]) -- transformed points
        """
        if self._coeffs is None:
            m = x.shape[0]
            out = np.zeros((m,self.dim_out))
            for i in range(m):
                cf = x[[i],self.dim_lin:]
                xx = x[i,:self.dim_lin]
                c = self._cMap.evaluate(cf)[0,:]
                T = self._TMap.evaluate(cf)[0,:,:]
                out[i,:] = c + np.dot(T, xx)
        else:
            xx = x[:,:self.dim_lin]
            out = self.c + np.dot(self.T, xx.T).T
        return out

    def grad_x(self, x, precomp=None, idxs_slice=slice(None)):
        if self._coeffs is None:
            m = x.shape[0]
            out = np.zeros((m,self.dim_out, self.dim_in))
            for i in range(m):
                cf = x[[i],self.dim_lin:]
                xx = x[i,:self.dim_lin]
                T = self._TMap.evaluate(cf)[0,:,:]
                gac = self._cMap.grad_x(cf)[0,:,:]
                gaT = self._TMap.grad_x(cf)[0,:,:,:]
                out[i,:,:self.dim_lin] = T
                out[i,:,self.dim_lin:] = gac + np.einsum('ijk,j->ik', gaT, xx)
        else:
            raise NotImplementedError("To be done")
        return out
        
class ConstantMap(Map):
    r""" Map :math:`T({\bf x})={\bf c}`

    Args:
       dim_in (int): input dimension :math:`d_x`
       const (:class:`ndarray<numpy.ndarray>`): constant :math:`{\bf c}`
    """
    def __init__(self, dim_in, const):
        self._const = const
        super(ConstantMap, self).__init__(dim_in, const.size)

    @property
    def const(self):
        return self._const
        
    def evaluate(self, x, precomp=None, idxs_slice=slice(None)):
        if x.shape[1] != self.dim_in:
            raise ValueError("dimension mismatch")
        tpl = tuple([nax] + [slice(None)] * self._const.ndim)
        return self._const[tpl]

    def grad_x(self, x, precomp=None, idxs_slice=slice(None)):
        if x.shape[1] != self.dim_in:
            raise ValueError("dimension mismatch")
        shp = (1,) + self._const.shape + (self.dim_in,)
        return np.zeros(shp)

    def hess_x(self, x, precomp=None, idxs_slice=slice(None)):
        if x.shape[1] != self.dim_in:
            raise ValueError("dimension mismatch")
        shp = (1,) + self._const.shape + (self.dim_in,self.dim_in)
        return np.zeros(shp)